#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <CL/cl.h>

#define WA 1024
#define HA 1024
#define WB 1024
#define HB WA
#define WC WB
#define HC HA



// Thread block size
#define BLOCK_SIZE 16

static inline void checkErr(cl_int err, cl_int success, const char * name) {
  if (err != success) {
    fprintf(stderr, "ERROR: %s\nErrorcode: %d\n", name, err);
    exit(EXIT_FAILURE);
  }
}

// Allocates a matrix with random float entries.
void randomInit(float* data, int size) {
  for (int i = 0; i < size; ++i)
    data[i] = rand() / (float)RAND_MAX;
}

//////////////////////////////////////////////////////////////////////////////
//! Loads a Program file.
//!
//! @return the source string if succeeded, 0 otherwise
//! @param cFilename        program filename
//! @param szFinalLength    returned length of the code string
//////////////////////////////////////////////////////////////////////////////
char* LoadProgSource(const char* cFilename, size_t* szFinalLength)
{
    // locals
    FILE* pFileStream = NULL;
    size_t szSourceLength;

    // open the OpenCL source code file
    #ifdef _WIN32   // Windows version
        if(fopen_s(&pFileStream, cFilename, "rb") != 0)
        {
            return NULL;
        }
    #else           // Linux version
        pFileStream = fopen(cFilename, "rb");
        if(pFileStream == 0)
        {
            return NULL;
        }
    #endif

    // get the length of the source code
    fseek(pFileStream, 0, SEEK_END);
    szSourceLength = ftell(pFileStream);
    fseek(pFileStream, 0, SEEK_SET);

    // allocate a buffer for the source code string and read it in
    char* cSourceString = (char *)malloc(szSourceLength + 1);
    if (fread((cSourceString), szSourceLength, 1, pFileStream) != 1)
    {
        fclose(pFileStream);
        free(cSourceString);
        return 0;
    }

    // close the file and return the total length of the combined (preamble + source) string
    fclose(pFileStream);
    if(szFinalLength != 0)
    {
        *szFinalLength = szSourceLength;
    }
    cSourceString[szSourceLength] = '\0';

    return cSourceString;
}

// Check bool
int isEqual(float a, float b) {
  return (fabs(a-b) < 0.001);
}

// Check Results

int checkResults(float* A, float* B, float* C) {
  unsigned int size_A = WA * HA;
  unsigned int size_B = WB * HB;
  unsigned int size_C = WC * HC;
  unsigned int bytesC = sizeof(float) * size_C;
  float* goldC = (float*) malloc(bytesC);
  for (int i=0; i < HC; i++) {
    for (int j=0; j < WC; j++) {
      goldC[i*WC + j] = 0;
      for (int k=0; k < HB; k++) {
        goldC[i*WC + j] += A[i*WA + k] * B[k*WB + j];
      }
      if(!isEqual(goldC[i*WC + j], C[i*WC + j])) {
        printf("Mismatch at %d,%d --- C = %f and goldC = %f\n", i, j, C[i*WC+j], goldC[i*WC+j]);
        return 0;
      }
    }
  }
  return 1; // Success
}

// GPU Computation of MatrixMul
void computeMatrixMul(float* h_A, unsigned bytes_A, float* h_B, unsigned bytes_B, float* h_C, unsigned bytes_C) {
 // OpenCL specific variables
  cl_context clGPUContext;
  cl_command_queue clCommandQue;
  cl_program clProgram;
  cl_kernel clKernel;

  size_t dataBytes;
  size_t kernelLength;
  cl_int errcode;

  // OpenCL device memory for matrices
  cl_mem d_A;
  cl_mem d_B;
  cl_mem d_C;

  /*****************************************/
  /* Initialize OpenCL */
  /*****************************************/
  // query the number of platforms
  cl_uint numPlatforms;
  errcode = clGetPlatformIDs(0, NULL, &numPlatforms);
  checkErr(errcode, CL_SUCCESS, "Failure to get number of platforms");

  // now get all the platform IDs
  cl_platform_id platforms[numPlatforms];
  errcode = clGetPlatformIDs(numPlatforms, platforms, NULL);
  checkErr(errcode, CL_SUCCESS, "Failure to get platform IDs");

  for(unsigned i=0; i < numPlatforms; i++) {
    char buffer[10240];
    //printf("  -- %d --\n", i);
    clGetPlatformInfo(platforms[i], CL_PLATFORM_PROFILE, 10240, buffer, NULL);
    //printf("  PROFILE = %s\n", buffer);
    clGetPlatformInfo(platforms[i], CL_PLATFORM_VERSION, 10240, buffer, NULL);
    //printf("  VERSION = %s\n", buffer);
    clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 10240, buffer, NULL);
    //printf("  NAME = %s\n", buffer);
    clGetPlatformInfo(platforms[i], CL_PLATFORM_VENDOR, 10240, buffer, NULL);
    //printf("  VENDOR = %s\n", buffer);
    clGetPlatformInfo(platforms[i], CL_PLATFORM_EXTENSIONS, 10240, buffer, NULL);
    //printf("  EXTENSIONS = %s\n", buffer);
  }
  // set platform property - just pick the first one
  cl_context_properties properties[] = {CL_CONTEXT_PLATFORM,
                                        (int) platforms[0],
                                        0};
  clGPUContext = clCreateContextFromType(properties, CL_DEVICE_TYPE_GPU,
                                         NULL, NULL, &errcode);
  checkErr(errcode, CL_SUCCESS, "Failure to create GPU context");

  // get the list of GPU devices associated with context
  errcode = clGetContextInfo(clGPUContext, CL_CONTEXT_DEVICES, 0,
                              NULL, &dataBytes);
  cl_device_id *clDevices = (cl_device_id *) malloc(dataBytes);
  errcode |= clGetContextInfo(clGPUContext, CL_CONTEXT_DEVICES, dataBytes,
                              clDevices, NULL);
  checkErr(errcode, CL_SUCCESS, "Failure to get context info");

  //Create a command-queue
  clCommandQue = clCreateCommandQueue(clGPUContext, clDevices[0], 0, &errcode);
  checkErr(errcode, CL_SUCCESS, "Failure to create command queue");

  // Setup device memory
  d_C = clCreateBuffer(clGPUContext, CL_MEM_READ_WRITE, bytes_C, NULL,
                       &errcode);
  d_A = clCreateBuffer(clGPUContext, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
                       bytes_A, h_A, &errcode);
  d_B = clCreateBuffer(clGPUContext, CL_MEM_READ_WRITE | CL_MEM_COPY_HOST_PTR,
                       bytes_B, h_B, &errcode);


   // Load and build OpenCL kernel
  /*char *clMatrixMul = LoadProgSource("matrixMul.cl",
                                        "// My comment\n",
                                        &kernelLength);*/
  //checkErr(clMatrixMul != NULL, 1 /*bool true*/, "Failure to load Program");

  /*clProgram = clCreateProgramWithSource(clGPUContext, 1,
                                        (const char **)&clMatrixMul,
                                        &kernelLength, &errcode);
  checkErr(errcode, CL_SUCCESS, "Failure to create program from source");
*/
  size_t binaryLength;
  char *clMatrixMul = LoadProgSource("matrixMul.nvptx.s", &binaryLength);
  checkErr(clMatrixMul != NULL, 1 /*bool true*/, "Failure to load Program Binary");
  
  cl_int binaryStatus;
  clProgram = clCreateProgramWithBinary(clGPUContext, 1, &clDevices[0],
                                        &binaryLength,
                                        (const unsigned char **)&clMatrixMul,
                                        &binaryStatus, &errcode);
  checkErr(errcode, CL_SUCCESS, "Failure to create program from binary");

  errcode = clBuildProgram(clProgram, 0, NULL, NULL, NULL, NULL);
  checkErr(errcode, CL_SUCCESS, "Failure to build program");

  clKernel = clCreateKernel(clProgram, "matrixMul", &errcode);
  checkErr(errcode, CL_SUCCESS, "Failure to create kernel");


  // Launch OpenCL kernel
  size_t localWorkSize[2], globalWorkSize[2];
 
  int wA = WA;
  int wC = WC;
  errcode = clSetKernelArg(clKernel, 0, sizeof(cl_mem), (void *)&d_C);
  errcode |= clSetKernelArg(clKernel, 1, sizeof(cl_mem), (void *)&d_A);
  errcode |= clSetKernelArg(clKernel, 2, sizeof(cl_mem), (void *)&d_B);
  errcode |= clSetKernelArg(clKernel, 3, sizeof(int), (void *)&wA);
  errcode |= clSetKernelArg(clKernel, 4, sizeof(int), (void *)&wC);
  checkErr(errcode, CL_SUCCESS, "Failure to set kernel arguments");
 
  localWorkSize[0] = BLOCK_SIZE;
  localWorkSize[1] = BLOCK_SIZE;
  globalWorkSize[0] = ((WB-1)/BLOCK_SIZE + 1) * BLOCK_SIZE;
  globalWorkSize[1] = ((HA-1)/BLOCK_SIZE + 1) * BLOCK_SIZE;
 
  errcode = clEnqueueNDRangeKernel(clCommandQue, clKernel, 2, NULL, 
                                   globalWorkSize, localWorkSize,
                                   0, NULL, NULL);
  checkErr(errcode, CL_SUCCESS, "Failure to enqueue kernel");
 
  // Retrieve result from device
  errcode = clEnqueueReadBuffer(clCommandQue, d_C, CL_TRUE, 0, bytes_C, 
                                h_C, 0, NULL, NULL);
  checkErr(errcode, CL_SUCCESS, "Failure to read buffer");
 
  // Print out the result
/*
  printf("\n\nMatrix C (Result)\n");
  for(int i = 0; i < size_C; i++) {
    printf("%f ", h_C[i]);
    if(((i + 1) % WC) == 0)
    printf("\n");
  }
  printf("\n");
 */
  // Deallocate memory

  clReleaseMemObject(d_A);
  clReleaseMemObject(d_C);
  clReleaseMemObject(d_B);

  free(clDevices);
  free(clMatrixMul);
  clReleaseContext(clGPUContext);
  clReleaseKernel(clKernel);
  clReleaseProgram(clProgram);
  clReleaseCommandQueue(clCommandQue);

}

// Main
int main(int argc, char** argv) {

  // seed for rand()
  srand(2006);

  // Allocate host memory for matrices A and B
  unsigned int size_A = WA * HA;
  unsigned int bytes_A = sizeof(float) * size_A;
  float* h_A = (float*) malloc(bytes_A);

  unsigned int size_B = WB * HB;
  unsigned int bytes_B = sizeof(float) * size_B;
  float* h_B = (float*) malloc(bytes_B);

   // Initialize host memory
   randomInit(h_A, size_A);
   randomInit(h_B, size_B);

/*
   // Print A and B
   printf("\n\nMatrix A\n");
   for(int i = 0; i < size_A; i++)
   {
      printf("%f ", h_A[i]);
      if(((i + 1) % WA) == 0)
      printf("\n");
   }

   printf("\n\nMatrix B\n");
   for(int i = 0; i < size_B; i++)
   {
      printf("%f ", h_B[i]);
      if(((i + 1) % WB) == 0)
      printf("\n");
   }
*/

  // Allocate host memory for the result matrix C
  unsigned int size_C = WC * HC;
  unsigned int bytes_C = sizeof(float) * size_C;
  float* h_C = (float*) malloc(bytes_C);

   // Compute using OpenCL
  computeMatrixMul(h_A, bytes_A, h_B, bytes_B, h_C, bytes_C);

  if(checkResults(h_A, h_B, h_C))
    printf("\nPass!\n");
  else
    printf("\nFailed!\n");
  printf("\nDone!\n");

  // Deallocate memory
  free(h_A);
  free(h_B);
  free(h_C);
}

