// Tile size
#define TILE_SIZE 16

// OpenCL Kernel for matrix multiply, C = A * B
// using shared memory and with boundary conditions
__kernel void matrixMul(__global float* C,
                        __global float* A,
                        __global float* B,
                        int m,
                        int k,
                        int n) {

  int tx = get_local_id(0); //2D Local Thread ID x
  int ty = get_local_id(1); //2D Local Thread ID y

  int bx = get_group_id(0); //2D Block ID x
  int by = get_group_id(1); //2D Block ID y

  int col = bx * TILE_SIZE + tx;
  int row = by * TILE_SIZE + ty;

  // Static work-group (thread block) local (shared) memory allocations
  __local float A_s[TILE_SIZE][TILE_SIZE];
  __local float B_s[TILE_SIZE][TILE_SIZE];

  // Loop counters
  int i, l;

  // Initialize accumulator
  float res = 0.0f;

  /* In the tiled version of matrix multiplication for arbitrary sizes,       * 
   * threads that are not matched with an element of the output matrix may    *
   * still participate in the memory transfer                                 */

  for (l = 0; l < k/TILE_SIZE; l++) {
    /* Loop for calculating with interior tiles (as far as k is concerned) *
     * Threads may still exceed the bottom limit of matrix A and the right *
     * limit of matrix B                                                   */

    /* check: thread does not exceed bottom limit of A */
    if (row < m)
        A_s[ty][tx] = A[row*k + TILE_SIZE*l + tx];

    /* check: thread does not exceed right limit of B */
    if (col < n)
        B_s[ty][tx] = B[(TILE_SIZE*l + ty)*n + col];

    /* Barrier: memory transfer must be completed before calculations start */
    barrier(CLK_LOCAL_MEM_FENCE);

    /* The threads that are within limits of output matrix accumulate in      *
     * their local result                                                     */
    if ((row < m) && (col < n))
      for (i = 0; i < TILE_SIZE; i++)
        res += A_s[ty][i]*B_s[i][tx];

    /* Barrier: calculations must be over before next memory transfer starts */
    barrier(CLK_LOCAL_MEM_FENCE);
  }

  /* For this last iteration, which will occur if matrices' sizes are not     *
   * proportional to the tile size, we must take care not to exceed the right *
   * limit of A and the bottom limit of B as well                             */

  /* 1st check: thread does not exceed right limit of A  *
   * 2st check: thread does not exceed bottom limit of A */
  if ((TILE_SIZE*l + tx < k) && (row < m))
    A_s[ty][tx] = A[row*k + TILE_SIZE*l + tx];

  /* 1st check: thread does not exceed bottom limit of B *
   * 2st check: thread does not exceed right limit of B  */
  if ((TILE_SIZE*l + ty < k) && (col < n))
    B_s[ty][tx] = B[(TILE_SIZE*l + ty)*n + col];

  /* Barrier: memory transfer must be completed before calculations can start */
  barrier(CLK_LOCAL_MEM_FENCE);

  /* The threads that are within limits of output matrix accumulate in their *
   * local result                                                            */
  if ((row < m) && (col < n)) {
    for (i = 0; i < min(TILE_SIZE, k - TILE_SIZE*l); i++)
      res += A_s[ty][i]*B_s[i][tx];
      /* Synchronization is not necessary, because each thread within limits *
       * of output matrix copies back its own private result                 */
      C[row*n + col] = res;
  }

  return;

}

