/***************************************************************************
 *cr
 *cr            (C) Copyright 2008-2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <parboil.h>

#include "atom.h"
#include "cutoff.h"
#include "output.h"
#include "macros.h"
#include <visc.h>

#define ERRTOL 1e-4f

#define NO_DEBUG
//#undef  NO_DEBUG

#define NOKERNELS             0
#define CUTOFF1               1
#define CUTOFF6              32
#define CUTOFF6OVERLAP       64
#define CUTOFFCPU         16384

#define mul24(x,y) (x)*(y)

// =================== CUTCP Graph =============================


// Define a type for a 3D coordinate.  Only 3 vector components are needed.
// Using int4 type because int3 support is missing on some platforms.
typedef struct __attribute__((__packed__)){
    int x;
    int y;
    int z;
    int w;
} xyz;
// May want to align these
typedef struct __attribute__((__packed__)) __attribute__((aligned(16))){
    float x;
    float y;
    float z;
    float w;
} float4;

void Allocation(long block) {
    // Memory shared between threadblocks
    size_t bytes_AtomBinCache = sizeof(float)*BIN_CACHE_MAXLEN * BIN_DEPTH * 4;
    void* AtomBinCache  = __visc__malloc(bytes_AtomBinCache);

    size_t bytes_myBinIndex = sizeof(xyz);
    void* myBinIndex = __visc__malloc(bytes_myBinIndex);
    __visc__return(4, AtomBinCache, bytes_AtomBinCache, myBinIndex, bytes_myBinIndex);
}

void CUTCPLeaf(
    int binDim_x,
    int binDim_y,
    float *binBaseAddr, size_t bytes_binBaseAddr,
    int offset,
    float h,                /* lattice spacing */
    float cutoff2,          /* square of cutoff distance */
    float inv_cutoff2,
    ener_t *regionZeroAddr, size_t bytes_regionZeroAddr, /* address of lattice regions starting at origin */
    int zRegionIndex,
    // constant memory arguments the next two
    int *NbrListLen, size_t bytes_NbrListLen,
    xyz *NbrList, size_t bytes_NbrList,
    // local memory args
    float* AtomBinCache, size_t bytes_AtomBinCache,
    int* myBinIndex, size_t bytes_myBinIndex
)
{
    __visc__hint(visc::DEVICE);
    __visc__attributes(4, binBaseAddr, regionZeroAddr, NbrListLen, NbrList, 1, regionZeroAddr);

    void* thisNode = __visc__getNode();
    void* parentNode = __visc__getParentNode(thisNode);
    int lx = __visc__getNodeInstanceID_x(thisNode);
    int ly = __visc__getNodeInstanceID_y(thisNode);
    int lz = __visc__getNodeInstanceID_z(thisNode);
    int gx = __visc__getNodeInstanceID_x(parentNode);
    int gy = __visc__getNodeInstanceID_y(parentNode);
    int dimx = __visc__getNumNodeInstances_x(thisNode);
    int dimy = __visc__getNumNodeInstances_y(thisNode);
    int gdimx = __visc__getNumNodeInstances_x(parentNode);
    int gdimy = __visc__getNumNodeInstances_y(parentNode);

    float* binZeroAddr = binBaseAddr + 4*offset;

    /*__local float AtomBinCache[BIN_CACHE_MAXLEN * BIN_DEPTH * 4];*/
    ener_t *myRegionAddr;
    /*__local xyz myBinIndex;*/

    const int xRegionIndex = gx;
    const int yRegionIndex = gy;

    /* thread id */
    const int tid = (lz*dimy+ly)*dimx+lx;

    /* neighbor index */
    int nbrid;

    /* this is the start of the sub-region indexed by tid */
    myRegionAddr = regionZeroAddr + ((zRegionIndex*gdimy
                                      + yRegionIndex)*gdimx + xRegionIndex)*REGION_SIZE;

    /* spatial coordinate of this lattice point */
    float x = (8 * xRegionIndex + lx) * h;
    float y = (8 * yRegionIndex + ly) * h;
    float z = (8 * zRegionIndex + lz) * h;

    int totalbins = 0;
    int numbins;

    /* bin number determined by center of region */
    myBinIndex[0] = (int) __visc__floor((8 * xRegionIndex + 4) * h * BIN_INVLEN);
    myBinIndex[1] = (int) __visc__floor((8 * yRegionIndex + 4) * h * BIN_INVLEN);
    myBinIndex[2] = (int) __visc__floor((8 * zRegionIndex + 4) * h * BIN_INVLEN);

    /* first neighbor in list for me to cache */
    nbrid = (tid >> 4);

    numbins = BIN_CACHE_MAXLEN;

#ifndef NEIGHBOR_COUNT
    ener_t energy0 = 0.f;
    ener_t energy1 = 0.f;
    ener_t energy2 = 0.f;
    ener_t energy3 = 0.f;
#else
    ener_t energy0 = 0, energy1 = 0, energy2 = 0, energy3 = 0;
#endif

    for (totalbins = 0;  totalbins < *NbrListLen;  totalbins += numbins) {

        int bincnt;

        /* start of where to write in shared memory */
        int startoff = BIN_SIZE * (tid >> 4);

        /* each half-warp to cache up to 4 atom bins */
        for (bincnt = 0;  bincnt < 4 && nbrid < *NbrListLen;  bincnt++, nbrid += 8) {

            int i = myBinIndex[0] + NbrList[nbrid].x;
            int j = myBinIndex[1] + NbrList[nbrid].y;
            int k = myBinIndex[2] + NbrList[nbrid].z;

            /* determine global memory location of atom bin */
            float *p_global = (( float *) binZeroAddr)
                              + (((mul24(k, binDim_y) + j)*binDim_x + i) << BIN_SHIFT);

            /* coalesced read from global memory -
             * retain same ordering in shared memory for now */
            int binIndex = startoff + (bincnt << (3 + BIN_SHIFT));
            int tidmask = tid & 15;

            AtomBinCache[binIndex + tidmask   ] = p_global[tidmask   ];
            AtomBinCache[binIndex + tidmask+16] = p_global[tidmask+16];
        }

        __visc__barrier();
        /* no warp divergence */
        if (totalbins + BIN_CACHE_MAXLEN > *NbrListLen) {
            numbins = *NbrListLen - totalbins;
        }

        int stopbin = (numbins << BIN_SHIFT);
        for (bincnt = 0; bincnt < stopbin; bincnt+=BIN_SIZE) {
            int i;

            for (i = 0;  i < BIN_DEPTH;  i++) {

                int off = bincnt + (i<<2);

                float aq = AtomBinCache[off + 3];
                if (0.f == aq)
                    break;  /* no more atoms in bin */

                float dx = AtomBinCache[off    ] - x;
                float dz = AtomBinCache[off + 2] - z;
                float dxdz2 = dx*dx + dz*dz;
                float dy = AtomBinCache[off + 1] - y;
                float r2 = dy*dy + dxdz2;

#ifndef NEIGHBOR_COUNT
                if (r2 < cutoff2)
                {
                    float s = (1.f - r2 * inv_cutoff2);
                    energy0 += aq * __visc__rsqrt(r2) * s * s;
                    //energy0 += aq * (1.0/__visc__sqrt(r2)) * s * s;
                }
#else
                energy0 += (r2 < cutoff2);
#endif
                dy -= 2.0f*h;
                r2 = dy*dy + dxdz2;

#ifndef NEIGHBOR_COUNT
                if (r2 < cutoff2)
                {
                    float s = (1.f - r2 * inv_cutoff2);
                    energy1 += aq * __visc__rsqrt(r2) * s * s;
                    //energy1 += aq * (1.0/__visc__sqrt(r2)) * s * s;
                }
#else
                energy1 += (r2 < cutoff2);
#endif
                dy -= 2.0f*h;
                r2 = dy*dy + dxdz2;
#ifndef NEIGHBOR_COUNT
                if (r2 < cutoff2)
                {
                    float s = (1.f - r2 * inv_cutoff2);
                    energy2 += aq * __visc__rsqrt(r2) * s * s;
                    //energy2 += aq * (1.0/__visc__sqrt(r2)) * s * s;
                }
#else
                energy2 += (r2 < cutoff2);
#endif
                dy -= 2.0f*h;
                r2 = dy*dy + dxdz2;
#ifndef NEIGHBOR_COUNT
                if (r2 < cutoff2)
                {
                    float s = (1.f - r2 * inv_cutoff2);
                    energy3 += aq * __visc__rsqrt(r2) * s * s;
                    //energy3 += aq * (1.0/__visc__rsqrt(r2)) * s * s;
                }
#else
                energy3 += (r2 < cutoff2);
#endif
            } /* end loop over atoms in bin */
        } /* end loop over cached atom bins */
        __visc__barrier();
    } /* end loop over neighbor list */

    /* store into global memory */
    myRegionAddr[(tid>>4)*64 + (tid&15)     ] = energy0;
    myRegionAddr[(tid>>4)*64 + (tid&15) + 16] = energy1;
    myRegionAddr[(tid>>4)*64 + (tid&15) + 32] = energy2;
    myRegionAddr[(tid>>4)*64 + (tid&15) + 48] = energy3;
}

void BlockingCUTCP(
    int binDim_x,
    int binDim_y,
    float4 *binBaseAddr, size_t bytes_binBaseAddr,
    int offset,
    float h,                /* lattice spacing */
    float cutoff2,          /* square of cutoff distance */
    float inv_cutoff2,
    ener_t *regionZeroAddr, size_t bytes_regionZeroAddr, /* address of lattice regions starting at origin */
    int zRegionIndex,
    // constant memory arguments the next two
    int *NbrListLen, size_t bytes_NbrListLen,
    xyz *NbrList, size_t bytes_NbrList,
    long blockx,
    long blocky,
    long blockz
) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(4, binBaseAddr, regionZeroAddr, NbrListLen, NbrList, 1, regionZeroAddr);

    void* AllocationNode = __visc__createNodeND(0, Allocation);
    void* CUTCPLeafNode = __visc__createNodeND(3, CUTCPLeaf, blockx, blocky, blockz);

    // Bind Inputs
    __visc__bindIn(AllocationNode, 15, 0, 0); // Bind blockx
    __visc__bindIn(CUTCPLeafNode, 0, 0, 0); // Bind binDim_x
    __visc__bindIn(CUTCPLeafNode, 1, 1, 0); // Bind binDim_y
    __visc__bindIn(CUTCPLeafNode, 2, 2, 0); // Bind binBaseAddr
    __visc__bindIn(CUTCPLeafNode, 3, 3, 0); // Bind bytes_binBaseAddr
    __visc__bindIn(CUTCPLeafNode, 4, 4, 0); // Bind offset
    __visc__bindIn(CUTCPLeafNode, 5, 5, 0); // Bind h
    __visc__bindIn(CUTCPLeafNode, 6, 6, 0); // Bind cutoff2
    __visc__bindIn(CUTCPLeafNode, 7, 7, 0); // Bind inv_cutoff2
    __visc__bindIn(CUTCPLeafNode, 8, 8, 0); // Bind regionZeroAddr
    __visc__bindIn(CUTCPLeafNode, 9, 9, 0); // Bind bytes_regionZeroAddr
    __visc__bindIn(CUTCPLeafNode, 10, 10, 0); // Bind zRegionIndex
    __visc__bindIn(CUTCPLeafNode, 11, 11, 0); // Bind NbrListLen
    __visc__bindIn(CUTCPLeafNode, 12, 12, 0); // Bind bytes_NbrListLen
    __visc__bindIn(CUTCPLeafNode, 13, 13, 0); // Bind NbrList
    __visc__bindIn(CUTCPLeafNode, 14, 14, 0); // Bind bytes_NbrList

    // Create Edges
    __visc__edge(AllocationNode, CUTCPLeafNode, 1, 0, 15, 0); // Edge AtomBinCache
    __visc__edge(AllocationNode, CUTCPLeafNode, 1, 1, 16, 0); // Edge bytes_AtomBinCache
    __visc__edge(AllocationNode, CUTCPLeafNode, 1, 2, 17, 0); // Edge myBinIndex
    __visc__edge(AllocationNode, CUTCPLeafNode, 1, 3, 18, 0); // Edge bytes_myBinIndex

}

typedef struct __attribute__((__packed__)) {
    int binDim_x;
    int binDim_y;
    float4 *binBaseAddr;
    size_t bytes_binBaseAddr;
    int offset;
    float h;                /* lattice spacing */
    float cutoff2;          /* square of cutoff distance */
    float inv_cutoff2;
    ener_t *regionZeroAddr;
    size_t bytes_regionZeroAddr; /* address of lattice regions starting at origin */
    int zRegionIndex;
    // constant memory arguments the next two
    int *NbrListLen;
    size_t bytes_NbrListLen;
    xyz *NbrList;
    size_t bytes_NbrList;
    long blockx;
    long blocky;
    long blockz;
    long gridx;
    long gridy;
    long gridz;
} RootIn;

void packData(
    RootIn* args,
    int binDim_x,
    int binDim_y,
    float4 *binBaseAddr, size_t bytes_binBaseAddr,
    int offset,
    float h,                /* lattice spacing */
    float cutoff2,          /* square of cutoff distance */
    float inv_cutoff2,
    ener_t *regionZeroAddr, size_t bytes_regionZeroAddr, /* address of lattice regions starting at origin */
    int zRegionIndex,
    // constant memory arguments the next two
    int *NbrListLen, size_t bytes_NbrListLen,
    xyz *NbrList, size_t bytes_NbrList,
    long blockx,
    long blocky,
    long blockz,
    long gridx,
    long gridy,
    long gridz
) {
    args->binDim_x = binDim_x;
    args->binDim_y = binDim_y;
    args->binBaseAddr = binBaseAddr;
    args->bytes_binBaseAddr = bytes_binBaseAddr;
    args->offset = offset;
    args->h = h;                /* lattice spacing */
    args->cutoff2 = cutoff2;          /* square of cutoff distance */
    args->inv_cutoff2 = inv_cutoff2;
    args->regionZeroAddr = regionZeroAddr;
    args->bytes_regionZeroAddr = bytes_regionZeroAddr; /* address of lattice regions starting at origin */
    args->zRegionIndex = zRegionIndex;
    // constant memory arguments the next two
    args->NbrListLen = NbrListLen;
    args->bytes_NbrListLen = bytes_NbrListLen;
    args->NbrList = NbrList;
    args->bytes_NbrList = bytes_NbrList;
    args->blockx = blockx;
    args->blocky = blocky;
    args->blockz = blockz;
    args->gridx = gridx;
    args->gridy = gridy;
    args->gridz = gridz;

}

void CUTCPRoot(
    int binDim_x,
    int binDim_y,
    float4 *binBaseAddr, size_t bytes_binBaseAddr,
    int offset,
    float h,                /* lattice spacing */
    float cutoff2,          /* square of cutoff distance */
    float inv_cutoff2,
    ener_t *regionZeroAddr, size_t bytes_regionZeroAddr, /* address of lattice regions starting at origin */
    int zRegionIndex,
    // constant memory arguments the next two
    int *NbrListLen, size_t bytes_NbrListLen,
    xyz *NbrList, size_t bytes_NbrList,
    long blockx,
    long blocky,
    long blockz,
    long gridx,
    long gridy,
    long gridz
) {
    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(4, binBaseAddr, regionZeroAddr, NbrListLen, NbrList, 1, regionZeroAddr);

    void* BlockingCUTCPNode = __visc__createNodeND(3, BlockingCUTCP, gridx, gridy, gridz);

    // Bind Inputs
    __visc__bindIn(BlockingCUTCPNode, 0, 0, 0); // Bind binDim_x
    __visc__bindIn(BlockingCUTCPNode, 1, 1, 0); // Bind binDim_y
    __visc__bindIn(BlockingCUTCPNode, 2, 2, 0); // Bind binBaseAddr
    __visc__bindIn(BlockingCUTCPNode, 3, 3, 0); // Bind bytes_binBaseAddr
    __visc__bindIn(BlockingCUTCPNode, 4, 4, 0); // Bind offset
    __visc__bindIn(BlockingCUTCPNode, 5, 5, 0); // Bind h
    __visc__bindIn(BlockingCUTCPNode, 6, 6, 0); // Bind cutoff2
    __visc__bindIn(BlockingCUTCPNode, 7, 7, 0); // Bind inv_cutoff2
    __visc__bindIn(BlockingCUTCPNode, 8, 8, 0); // Bind regionZeroAddr
    __visc__bindIn(BlockingCUTCPNode, 9, 9, 0); // Bind bytes_regionZeroAddr
    __visc__bindIn(BlockingCUTCPNode, 10, 10, 0); // Bind zRegionIndex
    __visc__bindIn(BlockingCUTCPNode, 11, 11, 0); // Bind NbrListLen
    __visc__bindIn(BlockingCUTCPNode, 12, 12, 0); // Bind bytes_NbrListLen
    __visc__bindIn(BlockingCUTCPNode, 13, 13, 0); // Bind NbrList
    __visc__bindIn(BlockingCUTCPNode, 14, 14, 0); // Bind bytes_NbrList
    __visc__bindIn(BlockingCUTCPNode, 15, 15, 0); // Bind blockx
    __visc__bindIn(BlockingCUTCPNode, 16, 16, 0); // Bind blocky
    __visc__bindIn(BlockingCUTCPNode, 17, 17, 0); // Bind blockz

}


// ==================== Host Code ==============================

int gpu_compute_cutoff_potential_lattice6overlap(
    struct pb_TimerSet *timers,        /* for measuring execution time */
    Lattice *lattice,
    float cutoff,                      /* cutoff distance */
    Atoms *atoms,                      /* array of atoms */
    int verbose                        /* print info/debug messages */
);

int appenddata(const char *filename, int size, double time) {
    FILE *fp;
    fp=fopen(filename, "a");
    if (fp == NULL) {
        printf("error appending to file %s..\n", filename);
        return -1;
    }
    fprintf(fp, "%d  %.3f\n", size, time);
    fclose(fp);
    return 0;
}

LatticeDim
lattice_from_bounding_box(Vec3 lo, Vec3 hi, float h)
{
    LatticeDim ret;

    ret.nx = (int) floorf((hi.x-lo.x)/h) + 1;
    ret.ny = (int) floorf((hi.y-lo.y)/h) + 1;
    ret.nz = (int) floorf((hi.z-lo.z)/h) + 1;
    ret.lo = lo;
    ret.h = h;

    return ret;
}

Lattice *
create_lattice(LatticeDim dim)
{
    int size;
    Lattice *lat = (Lattice *)malloc(sizeof(Lattice));

    if (lat == NULL) {
        fprintf(stderr, "Out of memory\n");
        exit(1);
    }

    lat->dim = dim;

    /* Round up the allocated size to a multiple of 8 */
    size = ((dim.nx * dim.ny * dim.nz) + 7) & ~7;
    lat->lattice = (float *)calloc(size, sizeof(float));

    if (lat->lattice == NULL) {
        fprintf(stderr, "Out of memory\n");
        exit(1);
    }

    return lat;
}


void
destroy_lattice(Lattice *lat)
{
    if (lat) {
        free(lat->lattice);
        free(lat);
    }
}

int main(int argc, char *argv[]) {
    Atoms *atom;

    LatticeDim lattice_dim;
    Lattice *gpu_lattice;
    Vec3 min_ext, max_ext;	/* Bounding box of atoms */
    Vec3 lo, hi;			/* Bounding box with padding  */

    float h = 0.5f;		/* Lattice spacing */
    float cutoff = 12.f;		/* Cutoff radius */
    float exclcutoff = 1.f;	/* Radius for exclusion */
    float padding = 0.5f;		/* Bounding box padding distance */

    int n;

    struct pb_Parameters *parameters;
    struct pb_TimerSet timers;

    /* Read input parameters */
    parameters = pb_ReadParameters(&argc, argv);
    if (parameters == NULL) {
        exit(1);
    }

    /* Expect one input file */
    if (pb_Parameters_CountInputs(parameters) != 1) {
        fprintf(stderr, "Expecting one input file\n");
        exit(1);
    }


    //pb_SwitchToTimer(&timers, pb_TimerID_IO);

    {
        const char *pqrfilename = parameters->inpFiles[0];

        if (!(atom = read_atom_file(pqrfilename))) {
            fprintf(stderr, "read_atom_file() failed\n");
            exit(1);
        }
        printf("read %d atoms from file '%s'\n", atom->size, pqrfilename);
    }

    pb_InitializeTimerSet(&timers);
    __visc__init();

    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

    /* find extent of domain */
    get_atom_extent(&min_ext, &max_ext, atom);
    printf("extent of domain is:\n");
    printf("  minimum %g %g %g\n", min_ext.x, min_ext.y, min_ext.z);
    printf("  maximum %g %g %g\n", max_ext.x, max_ext.y, max_ext.z);

    printf("padding domain by %g Angstroms\n", padding);
    lo = (Vec3) {
        min_ext.x - padding, min_ext.y - padding, min_ext.z - padding
    };
    hi = (Vec3) {
        max_ext.x + padding, max_ext.y + padding, max_ext.z + padding
    };
    printf("domain lengths are %g by %g by %g\n", hi.x-lo.x, hi.y-lo.y, hi.z-lo.z);

    lattice_dim = lattice_from_bounding_box(lo, hi, h);
    gpu_lattice = create_lattice(lattice_dim);

    /*
     *  OpenCL kernel, with overlapped GPU/CPU computation
     *  (Enter and exit the function with the COMPUTE timer active)
     */
    if (gpu_compute_cutoff_potential_lattice6overlap(&timers, gpu_lattice, cutoff, atom, 0)) {
        fprintf(stderr, "Computation failed\n");
        exit(1);
    }

    /*
     * Zero the lattice points that are too close to an atom.  This is
     * necessary for numerical stability.
     */
    if (remove_exclusions(gpu_lattice, exclcutoff, atom)) {
        fprintf(stderr, "remove_exclusions() failed for gpu lattice\n");
        exit(1);
    }

    pb_SwitchToTimer(&timers, pb_TimerID_NONE);
    pb_PrintTimerSet(&timers);
    __visc__cleanup();
    

    /* Print output */
    //pb_SwitchToTimer(&timers, pb_TimerID_IO);
    if (parameters->outFile) {
        write_lattice_summary(parameters->outFile, gpu_lattice);
    }
    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

    /* Cleanup */
    destroy_lattice(gpu_lattice);
    free_atom(atom);

    pb_FreeParameters(parameters);
    return 0;
}

int gpu_compute_cutoff_potential_lattice6overlap(
    struct pb_TimerSet *timers,        /* for measuring execution time */
    Lattice *lattice,
    float cutoff,                      /* cutoff distance */
    Atoms *atoms,                      /* array of atoms */
    int verbose                        /* print info/debug messages */
)
{
    int nx = lattice->dim.nx;
    int ny = lattice->dim.ny;
    int nz = lattice->dim.nz;
    float xlo = lattice->dim.lo.x;
    float ylo = lattice->dim.lo.y;
    float zlo = lattice->dim.lo.z;
    float h = lattice->dim.h;
    int natoms = atoms->size;
    Atom *atom = atoms->atoms;

    xyz nbrlist[NBRLIST_MAXLEN];
    size_t bytes_nbrlist = sizeof(xyz) * NBRLIST_MAXLEN;
    int* nbrlistlen = (int*) malloc(sizeof(int));

    int binHistoFull[BIN_DEPTH+1] = { 0 };   /* clear every array element */
    int binHistoCover[BIN_DEPTH+1] = { 0 };  /* clear every array element */
    int num_excluded = 0;

    int xRegionDim, yRegionDim, zRegionDim;
    int xRegionIndex, yRegionIndex, zRegionIndex;
    int xOffset, yOffset, zOffset;
    int lnx, lny, lnz, lnall;
    ener_t *regionZeroAddr, *thisRegion;
    int index, indexRegion;

    int c;
    xyz binDim;
    int nbins;
    float4 *binBaseAddr, *binZeroAddr;
    int *bincntBaseAddr, *bincntZeroAddr;
    Atoms *extra = NULL;

    int i, j, k, n;
    int sum, total;

    float avgFillFull, avgFillCover;
    const float cutoff2 = cutoff * cutoff;
    const float inv_cutoff2 = 1.f / cutoff2;

    long gridDim[3], blockDim[3];

#ifdef NEIGHBOR_COUNT
    double neighbor_count = 0;	/* used to profile the number of atoms near a
				 * lattice point */
#endif

    // Caller has made the "compute" timer active

    /* pad lattice to be factor of 8 in each dimension */
    xRegionDim = (int) ceilf(nx/8.f);
    yRegionDim = (int) ceilf(ny/8.f);
    zRegionDim = (int) ceilf(nz/8.f);

    lnx = 8 * xRegionDim;
    lny = 8 * yRegionDim;
    lnz = 8 * zRegionDim;
    lnall = lnx * lny * lnz;

    /* will receive energies from OpenCL */
    size_t bytes_regionZeroAddr = lnall * sizeof(ener_t);
    regionZeroAddr = (ener_t *) malloc(bytes_regionZeroAddr);

    /* create bins */
    c = (int) ceil(cutoff * BIN_INVLEN);  /* count extra bins around lattice */
    binDim.x = (int) ceil(lnx * h * BIN_INVLEN) + 2*c;
    binDim.y = (int) ceil(lny * h * BIN_INVLEN) + 2*c;
    binDim.z = (int) ceil(lnz * h * BIN_INVLEN) + 2*c;
    nbins = binDim.x * binDim.y * binDim.z;
    binBaseAddr = (float4 *) calloc(nbins * BIN_DEPTH, sizeof(float4));
    size_t bytes_binBaseAddr = nbins * BIN_DEPTH * sizeof(float4);

    binZeroAddr = binBaseAddr + ((c * binDim.y + c) * binDim.x + c) * BIN_DEPTH;

    bincntBaseAddr = (int *) calloc(nbins, sizeof(int));
    bincntZeroAddr = bincntBaseAddr + (c * binDim.y + c) * binDim.x + c;

    /* create neighbor list */
    if (ceilf(BIN_LENGTH / (8*h)) == floorf(BIN_LENGTH / (8*h))) {
        float s = sqrtf(3);
        float r2 = (cutoff + s*BIN_LENGTH) * (cutoff + s*BIN_LENGTH);
        int cnt = 0;
        /* develop neighbor list around 1 cell */
        if (2*c + 1 > NBRLIST_DIM) {
            fprintf(stderr, "must have cutoff <= %f\n",
                    (NBRLIST_DIM-1)/2 * BIN_LENGTH);
            return -1;
        }
        for (k = -c;  k <= c;  k++) {
            for (j = -c;  j <= c;  j++) {
                for (i = -c;  i <= c;  i++) {
                    if ((i*i + j*j + k*k)*BIN_LENGTH*BIN_LENGTH >= r2) continue;
                    nbrlist[cnt].x = i;
                    nbrlist[cnt].y = j;
                    nbrlist[cnt].z = k;
                    cnt++;
                }
            }
        }
        *nbrlistlen = cnt;
    }
    else if (8*h <= 2*BIN_LENGTH) {
        float s = 2.f*sqrtf(3);
        float r2 = (cutoff + s*BIN_LENGTH) * (cutoff + s*BIN_LENGTH);
        int cnt = 0;
        /* develop neighbor list around 3-cube of cells */
        if (2*c + 3 > NBRLIST_DIM) {
            fprintf(stderr, "must have cutoff <= %f\n",
                    (NBRLIST_DIM-3)/2 * BIN_LENGTH);
            return -1;
        }
        for (k = -c;  k <= c;  k++) {
            for (j = -c;  j <= c;  j++) {
                for (i = -c;  i <= c;  i++) {
                    if ((i*i + j*j + k*k)*BIN_LENGTH*BIN_LENGTH >= r2) continue;
                    nbrlist[cnt].x = i;
                    nbrlist[cnt].y = j;
                    nbrlist[cnt].z = k;
                    cnt++;
                }
            }
        }
        *nbrlistlen = cnt;
    }
    else {
        fprintf(stderr, "must have h <= %f\n", 0.25 * BIN_LENGTH);
        return -1;
    }

    /* perform geometric hashing of atoms into bins */
    {
        /* array of extra atoms, permit average of one extra per bin */
        Atom *extra_atoms = (Atom *) calloc(nbins, sizeof(Atom));
        int extra_len = 0;

        for (n = 0;  n < natoms;  n++) {
            float4 p;
            p.x = atom[n].x - xlo;
            p.y = atom[n].y - ylo;
            p.z = atom[n].z - zlo;
            p.w = atom[n].q;
            i = (int) floorf(p.x * BIN_INVLEN);
            j = (int) floorf(p.y * BIN_INVLEN);
            k = (int) floorf(p.z * BIN_INVLEN);
            if (i >= -c && i < binDim.x - c &&
                    j >= -c && j < binDim.y - c &&
                    k >= -c && k < binDim.z - c &&
                    atom[n].q != 0) {
                int index = (k * binDim.y + j) * binDim.x + i;
                float4 *bin = binZeroAddr + index * BIN_DEPTH;
                int bindex = bincntZeroAddr[index];
                if (bindex < BIN_DEPTH) {
                    /* copy atom into bin and increase counter for this bin */
                    bin[bindex] = p;
                    bincntZeroAddr[index]++;
                }
                else {
                    /* add index to array of extra atoms to be computed with CPU */
                    if (extra_len >= nbins) {
                        fprintf(stderr, "exceeded space for storing extra atoms\n");
                        return -1;
                    }
                    extra_atoms[extra_len] = atom[n];
                    extra_len++;
                }
            }
            else {
                /* excluded atoms are either outside bins or neutrally charged */
                num_excluded++;
            }
        }

        /* Save result */
        extra = (Atoms *)malloc(sizeof(Atoms));
        extra->atoms = extra_atoms;
        extra->size = extra_len;
    }

    /* bin stats */
    sum = total = 0;
    for (n = 0;  n < nbins;  n++) {
        binHistoFull[ bincntBaseAddr[n] ]++;
        sum += bincntBaseAddr[n];
        total += BIN_DEPTH;
    }
    avgFillFull = sum / (float) total;
    sum = total = 0;
    for (k = 0;  k < binDim.z - 2*c;  k++) {
        for (j = 0;  j < binDim.y - 2*c;  j++) {
            for (i = 0;  i < binDim.x - 2*c;  i++) {
                int index = (k * binDim.y + j) * binDim.x + i;
                binHistoCover[ bincntZeroAddr[index] ]++;
                sum += bincntZeroAddr[index];
                total += BIN_DEPTH;
            }
        }
    }
    avgFillCover = sum / (float) total;

    if (verbose) {
        /* report */
        printf("number of atoms = %d\n", natoms);
        printf("lattice spacing = %g\n", h);
        printf("cutoff distance = %g\n", cutoff);
        printf("\n");
        printf("requested lattice dimensions = %d %d %d\n", nx, ny, nz);
        printf("requested space dimensions = %g %g %g\n", nx*h, ny*h, nz*h);
        printf("expanded lattice dimensions = %d %d %d\n", lnx, lny, lnz);
        printf("expanded space dimensions = %g %g %g\n", lnx*h, lny*h, lnz*h);
        printf("number of bytes for lattice data = %lu\n", lnall*sizeof(float));
        printf("\n");
        printf("bin padding thickness = %d\n", c);
        printf("bin cover dimensions = %d %d %d\n",
               binDim.x - 2*c, binDim.y - 2*c, binDim.z - 2*c);
        printf("bin full dimensions = %d %d %d\n", binDim.x, binDim.y, binDim.z);
        printf("number of bins = %d\n", nbins);
        printf("total number of atom slots = %d\n", nbins * BIN_DEPTH);
        printf("%% overhead space = %g\n",
               (natoms / (double) (nbins * BIN_DEPTH)) * 100);
        printf("number of bytes for bin data = %lu\n",
               nbins * BIN_DEPTH * sizeof(float4));
        printf("\n");
        printf("bin histogram with padding:\n");
        sum = 0;
        for (n = 0;  n <= BIN_DEPTH;  n++) {
            printf("     number of bins with %d atoms:  %d\n", n, binHistoFull[n]);
            sum += binHistoFull[n];
        }
        printf("     total number of bins:  %d\n", sum);
        printf("     %% average fill:  %g\n", avgFillFull * 100);
        printf("\n");
        printf("bin histogram excluding padding:\n");
        sum = 0;
        for (n = 0;  n <= BIN_DEPTH;  n++) {
            printf("     number of bins with %d atoms:  %d\n", n, binHistoCover[n]);
            sum += binHistoCover[n];
        }
        printf("     total number of bins:  %d\n", sum);
        printf("     %% average fill:  %g\n", avgFillCover * 100);
        printf("\n");
        printf("number of extra atoms = %d\n", extra->size);
        printf("%% atoms that are extra = %g\n", (extra->size / (double) natoms) * 100);
        printf("\n");

        /* sanity check on bins */
        sum = 0;
        for (n = 0;  n <= BIN_DEPTH;  n++) {
            sum += n * binHistoFull[n];
        }
        sum += extra->size + num_excluded;
        printf("sanity check on bin histogram with edges:  "
               "sum + others = %d\n", sum);
        sum = 0;
        for (n = 0;  n <= BIN_DEPTH;  n++) {
            sum += n * binHistoCover[n];
        }
        sum += extra->size + num_excluded;
        printf("sanity check on bin histogram excluding edges:  "
               "sum + others = %d\n", sum);
        printf("\n");

        /* neighbor list */
        printf("neighbor list length = %d\n", *nbrlistlen);
        printf("\n");
    }

    // Track visc data
    llvm_visc_track_mem(regionZeroAddr, bytes_regionZeroAddr);
    llvm_visc_track_mem(binBaseAddr, bytes_binBaseAddr);
    llvm_visc_track_mem(nbrlistlen, sizeof(int));
    llvm_visc_track_mem(nbrlist, bytes_nbrlist);

    /* setup OpenCL kernel parameters */
    blockDim[0] = 8;
    blockDim[1] = 2;
    blockDim[2] = 8;
    gridDim[0] = xRegionDim;
    gridDim[1] = yRegionDim;
    gridDim[2] = 1;

    /* allocate and initialize memory on OpenCL device */
    if (verbose) {
        printf("Allocating %.2fMB on OpenCL device for potentials\n",
               lnall * sizeof(float) / (double) (1024*1024));
    }

    memset(regionZeroAddr,0,lnall*sizeof(ener_t));

    if (verbose) {
        printf("Allocating %.2fMB on OpenCL device for atom bins\n",
               nbins * BIN_DEPTH * sizeof(float4) / (double) (1024*1024));
    }

    //Sub buffers are not supported in OpenCL v1.0
    int offset = ((c * binDim.y + c) * binDim.x + c) * BIN_DEPTH;

    if (verbose)
        printf("\n");


    RootIn* args = (RootIn*) malloc(sizeof(RootIn));
    packData( args,
              binDim.x,
              binDim.y,
              binBaseAddr,
              bytes_binBaseAddr,
              offset,
              h,
              cutoff2,
              inv_cutoff2,
              regionZeroAddr,
              bytes_regionZeroAddr,
              zRegionIndex,
              nbrlistlen,
              (size_t )sizeof(int),
              nbrlist,
              bytes_nbrlist,
              blockDim[0],
              blockDim[1],
              blockDim[2],
              gridDim[0],
              gridDim[1],
              gridDim[2]
            );

    /* loop over z-dimension, invoke OpenCL kernel for each x-y plane */
    pb_SwitchToTimer(timers, visc_TimerID_COMPUTATION);
    void* CUTCP_DFG;
    if(verbose)
      printf("Invoking OpenCL kernel on %d region planes...\n", zRegionDim);
    for (zRegionIndex = 0;  zRegionIndex < zRegionDim;  zRegionIndex++) {
#ifndef NO_DEBUG
        printf("  computing plane %d\n", zRegionIndex);
        fflush(stdout);
#endif

        args->zRegionIndex = zRegionIndex;

        CUTCP_DFG = __visc__launch(0, CUTCPRoot, (void*)args);
        __visc__wait(CUTCP_DFG);
        //llvm_visc_request_mem(regionZeroAddr, lnall*sizeof(ener_t));
    }

    /*
     * handle extra atoms on the CPU, concurrently with the GPU calculations
     */

    pb_SwitchToTimer(timers, pb_TimerID_COMPUTE);
    if (extra->size > 0) {
        if(verbose) {
          printf("computing extra atoms on CPU\n");
        }

        pb_SwitchToTimer(timers, visc_TimerID_MISC);

        if (cpu_compute_cutoff_potential_lattice(lattice, cutoff, extra)) {
            fprintf(stderr, "cpu_compute_cutoff_potential_lattice() failed "
                    "for extra atoms\n");
            return -1;
        }
        pb_SwitchToTimer(timers, visc_TimerID_MISC);
        printf("\n");
    }
    if(verbose)
      printf("Finished OpenCL kernel calls                        \n");

    /* copy result regions from OpenCL device */
    pb_SwitchToTimer(timers, pb_TimerID_COPY);

    llvm_visc_request_mem(regionZeroAddr, lnall*sizeof(ener_t));

    /*
     * transpose on CPU, updating, producing the final lattice
     */
    /* transpose regions back into lattice */
    pb_SwitchToTimer(timers, pb_TimerID_COMPUTE);
    for (k = 0;  k < nz;  k++) {
        zRegionIndex = (k >> 3);
        zOffset = (k & 7);

        for (j = 0;  j < ny;  j++) {
            yRegionIndex = (j >> 3);
            yOffset = (j & 7);

            for (i = 0;  i < nx;  i++) {
                xRegionIndex = (i >> 3);
                xOffset = (i & 7);

                thisRegion = regionZeroAddr
                             + ((zRegionIndex * yRegionDim + yRegionIndex) * xRegionDim
                                + xRegionIndex) * REGION_SIZE;

                indexRegion = (zOffset * 8 + yOffset) * 8 + xOffset;
                index = (k * ny + j) * nx + i;

#ifndef NEIGHBOR_COUNT
                lattice->lattice[index] += thisRegion[indexRegion];
#else
                neighbor_count += thisRegion[indexRegion];
#endif
            }
        }
    }

#ifdef NEIGHBOR_COUNT
    printf("Neighbor count: %f\n", (float)neighbor_count);
#endif

    /* cleanup memory allocations */
    free(regionZeroAddr);
    free(binBaseAddr);
    free(bincntBaseAddr);
    free_atom(extra);

    return 0;
}
