/***************************************************************************
 *
 *            (C) Copyright 2010 The Board of Trustees of the
 *                        University of Illinois
 *                         All Rights Reserved
 *
 ***************************************************************************/


#include <parboil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <visc.h>
#include "util.h"

#define BLOCK_X             14

#define PRESCAN_THREADS     512
#define PRESCAN_BLOCKS_X    64

#define UNROLL              16

#define KB                  24
#define BINS_PER_BLOCK      ((KB)*1024)

#define min(x,y)            ((x) < (y))? (x) : (y)

#define debug
//#undef debug

/******************************************************************************
* Implementation: GPU
* Details:
* in the GPU implementation of histogram, we begin by computing the span of the
* input values into the histogram. Then the histogramming computation is carried
* out by a (BLOCK_X, BLOCK_Y) sized grid, where every group of Y (same X)
* computes its own partial histogram for a part of the input, and every Y in the
* group exclusively writes to a portion of the span computed in the beginning.
* Finally, a reduction is performed to combine all the partial histograms into
* the final result.
******************************************************************************/
// Prescan

void PrescanAllocation (int block) {
    // Memory shared between threadblocks
    size_t bytes_Avg = sizeof(float)*PRESCAN_THREADS;
    void* Avg  = __visc__malloc(bytes_Avg);

    size_t bytes_StdDev = sizeof(float)*PRESCAN_THREADS;
    void* StdDev = __visc__malloc(bytes_StdDev);
    __visc__return(Avg, bytes_Avg, StdDev, bytes_StdDev);
}

void PrescanLeaf (unsigned int* input, size_t bytes_input,
                  int size,
                  unsigned int* minmax, size_t bytes_minmax,
                  // local memory arguments
                  float* Avg, size_t bytes_Avg,
                  float* StdDev, size_t bytes_StdDev) {


    __visc__hint(visc::DEVICE);
    __visc__attributes(2, input, minmax, 1, minmax);

    void* thisNode = __visc__getNode();
    void* parentNode = __visc__getParentNode(thisNode);
    int lx = __visc__getNodeInstanceID_x(thisNode);
    int gx = __visc__getNodeInstanceID_x(parentNode);
    int dimx = __visc__getNumNodeInstances_x(thisNode);
    int gdimx = __visc__getNumNodeInstances_x(parentNode);

    int threadIdxx = lx;
    int blockDimx = dimx;
    int blockIdxx = gx;
    int stride = size/gdimx;
    int addr = blockIdxx*stride+threadIdxx;
    int end = blockIdxx*stride + stride/8; // Only sample 1/8th of the input data

    // Compute the average per thread
    float avg = 0.0;
    unsigned int count = 0;
    while (addr < end) {
        avg += input[addr];
        count++;
        addr += blockDimx;
    }
    avg /= count;
    Avg[threadIdxx] = avg;

    // Compute the standard deviation per thread
    int addr2 = blockIdxx*stride+threadIdxx;
    float stddev = 0;
    while (addr2 < end) {
        stddev += (input[addr2]-avg)*(input[addr2]-avg);
        addr2 += blockDimx;
    }
    stddev /= count;
    StdDev[threadIdxx] = __visc__sqrt(stddev);

#define SUM(stride__)\
if(threadIdxx < stride__){\
    Avg[threadIdxx] += Avg[threadIdxx+stride__];\
    StdDev[threadIdxx] += StdDev[threadIdxx+stride__];\
}

    // Add all the averages and standard deviations from all the threads
    // and take their arithmetic average (as a simplified approximation of the
    // real average and standard deviation.
#if (PRESCAN_THREADS >= 32)
    for (int stride = PRESCAN_THREADS/2; stride >= 32; stride = stride >> 1) {
        __visc__barrier();
        SUM(stride);
    }
#endif
#if (PRESCAN_THREADS >= 16)
    SUM(16);
#endif
#if (PRESCAN_THREADS >= 8)
    SUM(8);
#endif
#if (PRESCAN_THREADS >= 4)
    SUM(4);
#endif
#if (PRESCAN_THREADS >= 2)
    SUM(2);
#endif

    if (threadIdxx == 0) {
        float avg = Avg[0]+Avg[1];
        avg /= PRESCAN_THREADS;
        float stddev = StdDev[0]+StdDev[1];
        stddev /= PRESCAN_THREADS;

        // Take the maximum and minimum range from all the blocks. This will
        // be the final answer. The standard deviation is taken out to 10 sigma
        // away from the average. The value 10 was obtained empirically.
        __visc__atomic_umin((int*)minmax,((unsigned int)(avg-10*stddev))/(KB*1024));
        __visc__atomic_umax((int*)minmax+1,((unsigned int)(avg+10*stddev))/(KB*1024));
    }
}

void PrescanBlock (unsigned int* input, size_t bytes_input,
                   int size,
                   unsigned int* minmax, size_t bytes_minmax,
                   int block) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(2, input, minmax, 1, minmax);

    void* AllocationNode = __visc__createNode(PrescanAllocation);
    void* PrescanLeafNode = __visc__createNode1D(PrescanLeaf, block);

    // Bind Inputs
    __visc__bindIn(AllocationNode, 5, 0, 0);  // Bind block

    __visc__bindIn(PrescanLeafNode, 0, 0, 0); // Bind input
    __visc__bindIn(PrescanLeafNode, 1, 1, 0); // Bind bytes_input
    __visc__bindIn(PrescanLeafNode, 2, 2, 0); // Bind size
    __visc__bindIn(PrescanLeafNode, 3, 3, 0); // Bind minmax
    __visc__bindIn(PrescanLeafNode, 4, 4, 0); // Bind bytes_minmax

    // Create Edges
    __visc__edge(AllocationNode, PrescanLeafNode, 1, 0, 5, 0); // Edge Avg
    __visc__edge(AllocationNode, PrescanLeafNode, 1, 1, 6, 0); // Edge bytes_Avg
    __visc__edge(AllocationNode, PrescanLeafNode, 1, 2, 7, 0); // Edge StdDev
    __visc__edge(AllocationNode, PrescanLeafNode, 1, 3, 8, 0); // Edge bytes_StdDev

}

void PrescanRoot (unsigned int* input, size_t bytes_input,
                  int size,
                  unsigned int* minmax, size_t bytes_minmax,
                  int block,
                  int grid) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(2, input, minmax, 1, minmax);
    //__visc__attributes(1, input, 1, minmax);

    void* PrescanBlockNode = __visc__createNode1D(PrescanBlock, grid);

    // Bind Inputs
    __visc__bindIn(PrescanBlockNode, 0, 0, 0); // Bind input
    __visc__bindIn(PrescanBlockNode, 1, 1, 0); // Bind bytes_input
    __visc__bindIn(PrescanBlockNode, 2, 2, 0); // Bind size
    __visc__bindIn(PrescanBlockNode, 3, 3, 0); // Bind minmax
    __visc__bindIn(PrescanBlockNode, 4, 4, 0); // Bind bytes_minmax
    __visc__bindIn(PrescanBlockNode, 5, 5, 0); // Bind block

}

typedef struct __attribute__((__packed__)) {
    unsigned int* input;
    size_t bytes_input;
    int size;
    unsigned int* minmax;
    size_t bytes_minmax;
    int block;
    int grid;
}
PrescanRootIn;

void PrescanPackData( PrescanRootIn* args,
                      unsigned int* input, size_t bytes_input,
                      int size,
                      unsigned int* minmax, size_t bytes_minmax,
                      int block,
                      int grid
                    ) {
    args->input = input;
    args->bytes_input = bytes_input;
    args->size = size;
    args->minmax = minmax;
    args->bytes_minmax = bytes_minmax;
    args->block = block;
    args->grid = grid;
}

// Intermediates
typedef struct __attribute__((__packed__)) {
    unsigned char x;
    unsigned char y;
    unsigned char z;
    unsigned char w;
}
uchar4;

typedef struct __attribute__((__packed__)) {
    unsigned int x;
    unsigned int y;
}
uint2;

__attribute__((always_inline))
static inline void calculateBin (unsigned int bin, uchar4 *sm_mapping) {

    unsigned char offset  =  bin        %   4;
    unsigned char indexlo = (bin >>  2) % 256;
    unsigned char indexhi = (bin >> 10) %  KB;
    unsigned char block   =  bin / BINS_PER_BLOCK;

    offset *= 8;

    uchar4 sm;
    sm.x = block;
    sm.y = indexhi;
    sm.z = indexlo;
    sm.w = offset;

    *sm_mapping = sm;
}

void IntermediatesLeaf( uint2 *input, size_t bytes_input,
                        unsigned int height,
                        unsigned int width,
                        unsigned int input_pitch,
                        uchar4 *sm_mappings, size_t bytes_sm_mappings) {

    __visc__hint(visc::DEVICE);
    __visc__attributes(1, input, 1, sm_mappings);

    void* thisNode = __visc__getNode();
    void* parentNode = __visc__getParentNode(thisNode);
    int lx = __visc__getNodeInstanceID_x(thisNode);
    int gx = __visc__getNodeInstanceID_x(parentNode);
    int dimx = __visc__getNumNodeInstances_x(thisNode);

    int threadIdxx = lx;
    int blockDimx = dimx;
    unsigned int line = UNROLL * (gx);// 16 is the unroll factor;

    uint2 *load_bin = input + line * input_pitch + threadIdxx;

    unsigned int store = line * width + threadIdxx;
    bool skip = (width % 2) && (threadIdxx == (blockDimx - 1));

#pragma unroll
    for (int i = 0; i < UNROLL; i++)
    {
        uint2 bin_value = *load_bin;

        calculateBin (
            bin_value.x,
            &sm_mappings[store]
        );

        if (!skip) calculateBin (
                bin_value.y,
                &sm_mappings[store + blockDimx]
            );

        load_bin += input_pitch;
        store += width;
    }
}

void IntermediatesBlock(uint2 *input, size_t bytes_input,
                        unsigned int height,
                        unsigned int width,
                        unsigned int input_pitch,
                        uchar4 *sm_mappings, size_t bytes_sm_mappings,
                        int block) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(1, input, 1, sm_mappings);

    void* IntermediatesLeafNode = __visc__createNode1D(IntermediatesLeaf, block);

    // Bind Inputs
    __visc__bindIn(IntermediatesLeafNode, 0, 0, 0); // Bind Input
    __visc__bindIn(IntermediatesLeafNode, 1, 1, 0); // Bind bytes_input
    __visc__bindIn(IntermediatesLeafNode, 2, 2, 0); // Bind height
    __visc__bindIn(IntermediatesLeafNode, 3, 3, 0); // Bind width
    __visc__bindIn(IntermediatesLeafNode, 4, 4, 0); // Bind input_pitch
    __visc__bindIn(IntermediatesLeafNode, 5, 5, 0); // Bind sm_mappings
    __visc__bindIn(IntermediatesLeafNode, 6, 6, 0); // Bind bytes_sm_mappings
}

void IntermediatesRoot( uint2 *input, size_t bytes_input,
                        unsigned int height,
                        unsigned int width,
                        unsigned int input_pitch,
                        uchar4 *sm_mappings, size_t bytes_sm_mappings,
                        int block,
                        int grid) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(1, input, 1, sm_mappings);

    void* IntermediatesBlockNode = __visc__createNode1D(IntermediatesBlock, grid);

    // Bind Inputs
    __visc__bindIn(IntermediatesBlockNode, 0, 0, 0); // Bind Input
    __visc__bindIn(IntermediatesBlockNode, 1, 1, 0); // Bind bytes_input
    __visc__bindIn(IntermediatesBlockNode, 2, 2, 0); // Bind height
    __visc__bindIn(IntermediatesBlockNode, 3, 3, 0); // Bind width
    __visc__bindIn(IntermediatesBlockNode, 4, 4, 0); // Bind input_pitch
    __visc__bindIn(IntermediatesBlockNode, 5, 5, 0); // Bind sm_mappings
    __visc__bindIn(IntermediatesBlockNode, 6, 6, 0); // Bind bytes_sm_mappings
    __visc__bindIn(IntermediatesBlockNode, 7, 7, 0); // Bind block
}

typedef struct __attribute__((__packed__)) {
    unsigned int* input;
    size_t bytes_input;
    unsigned height;
    unsigned width;
    unsigned input_pitch;
    unsigned char* sm_mappings;
    size_t bytes_sm_mappings;
    int block;
    int grid;
}
IntermediatesRootIn;

void IntermediatesPackData( IntermediatesRootIn* args,
                            unsigned int* input, size_t bytes_input,
                            unsigned height,
                            unsigned width,
                            unsigned input_pitch,
                            unsigned char* sm_mappings,
                            size_t bytes_sm_mappings,
                            int block,
                            int grid
                          ) {
    args->input = input;
    args->bytes_input = bytes_input;
    args->height = height;
    args->width = width;
    args->input_pitch = input_pitch;
    args->sm_mappings = sm_mappings;
    args->bytes_sm_mappings = bytes_sm_mappings;
    args->block = block;
    args->grid = grid;
}

// Histo Main

__attribute__((always_inline))
static inline void testIncrementGlobal (
    unsigned int *global_histo,
    unsigned int sm_range_min,
    unsigned int sm_range_max,
    const uchar4 sm)
{
    const unsigned int range = sm.x;
    const unsigned int indexhi = sm.y;
    const unsigned int indexlo = sm.z;
    const unsigned int offset  = sm.w;

    /* Scan for inputs that are outside the central region of histogram */
    if (range < sm_range_min || range > sm_range_max)
    {
        const unsigned int bin = range * BINS_PER_BLOCK + offset / 8 + (indexlo << 2) + (indexhi << 10);
        const unsigned int bin_div2 = bin / 2;
        const unsigned int bin_offset = (bin % 2 == 1) ? 16 : 0;

        unsigned int old_val = global_histo[bin_div2];
        unsigned short old_bin = (old_val >> bin_offset) & 0xFFFF;

        if (old_bin < 255)
        {
            __visc__atomic_add ((int*)&global_histo[bin_div2], 1 << bin_offset);
        }
    }
}

#define smem(x,y) smem[(x)*256+(y)]

__attribute__((always_inline))
static inline void testIncrementLocal (
    unsigned int *global_overflow,
    unsigned int* smem,
    const unsigned int myRange,
    const uchar4 sm)
{
    const unsigned int range = sm.x;
    const unsigned int indexhi = sm.y;
    const unsigned int indexlo = sm.z;
    const unsigned int offset  = sm.w;

    /* Scan for inputs that are inside the central region of histogram */
    if (range == myRange)
    {
        /* Atomically increment shared memory */
        unsigned int add = (unsigned int)(1 << offset);
        unsigned int prev = __visc__atomic_add ((int*)&smem(indexhi, indexlo), add);

        /* Check if current bin overflowed */
        unsigned int prev_bin_val = (prev >> offset) & 0x000000FF;

        /* If there was an overflow, record it and record if it cascaded into other bins */
        if (prev_bin_val == 0x000000FF)
        {
            const unsigned int bin =
                range * BINS_PER_BLOCK +
                offset / 8 + (indexlo << 2) + (indexhi << 10);

            bool can_overflow_to_bin_plus_1 = (offset < 24) ? true : false;
            bool can_overflow_to_bin_plus_2 = (offset < 16) ? true : false;
            bool can_overflow_to_bin_plus_3 = (offset <  8) ? true : false;

            bool overflow_into_bin_plus_1 = false;
            bool overflow_into_bin_plus_2 = false;
            bool overflow_into_bin_plus_3 = false;

            unsigned int prev_bin_plus_1_val = (prev >> (offset +  8)) & 0x000000FF;
            unsigned int prev_bin_plus_2_val = (prev >> (offset + 16)) & 0x000000FF;
            unsigned int prev_bin_plus_3_val = (prev >> (offset + 24)) & 0x000000FF;

            if (can_overflow_to_bin_plus_1 &&        prev_bin_val == 0x000000FF) overflow_into_bin_plus_1 = true;
            if (can_overflow_to_bin_plus_2 && prev_bin_plus_1_val == 0x000000FF) overflow_into_bin_plus_2 = true;
            if (can_overflow_to_bin_plus_3 && prev_bin_plus_2_val == 0x000000FF) overflow_into_bin_plus_3 = true;

            unsigned int bin_plus_1_add;
            unsigned int bin_plus_2_add;
            unsigned int bin_plus_3_add;

            if (overflow_into_bin_plus_1) bin_plus_1_add = (prev_bin_plus_1_val < 0x000000FF) ? 0xFFFFFFFF : 0x000000FF;
            if (overflow_into_bin_plus_2) bin_plus_2_add = (prev_bin_plus_2_val < 0x000000FF) ? 0xFFFFFFFF : 0x000000FF;
            if (overflow_into_bin_plus_3) bin_plus_3_add = (prev_bin_plus_3_val < 0x000000FF) ? 0xFFFFFFFF : 0x000000FF;

            __visc__atomic_add ((int*)&global_overflow[bin],   256);
            if (overflow_into_bin_plus_1) __visc__atomic_add ((int*)&global_overflow[bin+1], bin_plus_1_add);
            if (overflow_into_bin_plus_2) __visc__atomic_add ((int*)&global_overflow[bin+2], bin_plus_2_add);
            if (overflow_into_bin_plus_3) __visc__atomic_add ((int*)&global_overflow[bin+3], bin_plus_3_add);
        }
    }
}

__attribute__((always_inline))
static inline void clearMemory (unsigned int* smem, int lx, int dimx)
{
    for (int i = lx, blockDimx = dimx; i < BINS_PER_BLOCK / 4; i += blockDimx)
    {
        ((unsigned int*)smem)[i] = 0;
    }
}

__attribute__((always_inline))
static inline void copyMemory (unsigned int *dst, unsigned int* src, int lx, int dimx)
{
    for (int i = lx, blockDimx = dimx; i < BINS_PER_BLOCK/4; i += blockDimx)
    {
        dst[i] = ((unsigned int*)src)[i];
    }
}

void MainAllocation (int block) {
    // Memory shared between threadblocks
    size_t bytes_sub_histo = sizeof(unsigned int) * KB * 256;
    void* sub_histo  = __visc__malloc(bytes_sub_histo);

    __visc__return(sub_histo, bytes_sub_histo);
}

void MainLeaf ( uchar4* sm_mappings, size_t bytes_sm_mappings,
                unsigned int num_elements,
                unsigned int sm_range_min,
                unsigned int sm_range_max,
                unsigned int histo_height,
                unsigned int histo_width,
                unsigned int* global_subhisto, size_t bytes_global_subhisto,
                unsigned int* global_histo, size_t bytes_global_histo,
                unsigned int* global_overflow, size_t bytes_global_overflow,
                // local memory
                unsigned int* sub_histo, size_t bytes_sub_histo) {
    /* Most optimal solution uses 24 * 1024 bins per threadblock */
    //__local unsigned int sub_histo[KB][256];

    /* Each threadblock contributes to a specific 24KB range of histogram,
     * and also scans every N-th line for interesting data.  N = gridDim.x
     */

    __visc__hint(visc::DEVICE);
    __visc__attributes(1, sm_mappings, 3, global_subhisto, global_histo, global_overflow);

    void* thisNode = __visc__getNode();
    void* parentNode = __visc__getParentNode(thisNode);

    int lx = __visc__getNodeInstanceID_x(thisNode);
    int gx = __visc__getNodeInstanceID_x(parentNode);
    int dimx = __visc__getNumNodeInstances_x(thisNode);
    int gdimx = __visc__getNumNodeInstances_x(parentNode);

    int gy = __visc__getNodeInstanceID_y(parentNode);

    unsigned int blockDimx = dimx;
    unsigned int gridDimx = gdimx;
    unsigned int local_scan_range = sm_range_min + gy;
    unsigned int local_scan_load = gx * blockDimx + lx;

    clearMemory (sub_histo, lx, dimx);
    __visc__barrier();

    if (gy == 0)
    {
        /* Loop through and scan the input */
        while (local_scan_load < num_elements)
        {
            /* Read buffer */
            uchar4 sm = sm_mappings[local_scan_load];
            local_scan_load += blockDimx * gridDimx;

            /* Check input */
            testIncrementLocal (
                global_overflow,
                sub_histo,
                local_scan_range,
                sm
            );
            testIncrementGlobal (
                global_histo,
                sm_range_min,
                sm_range_max,
                sm
            );
        }
    }
    else
    {
        /* Loop through and scan the input */
        while (local_scan_load < num_elements)
        {
            /* Read buffer */
            uchar4 sm = sm_mappings[local_scan_load];
            local_scan_load += blockDimx * gridDimx;

            /* Check input */
            testIncrementLocal (
                global_overflow,
                sub_histo,
                local_scan_range,
                sm
            );
        }
    }

    /* Store sub histogram to global memory */
    unsigned int store_index = gx * (histo_height * histo_width / 4) + (local_scan_range * BINS_PER_BLOCK / 4);//(local_scan_range * BINS_PER_BLOCK);

    __visc__barrier();
    copyMemory (&(global_subhisto[store_index]), sub_histo, lx, dimx);
}

void MainBlock (uchar4* sm_mappings, size_t bytes_sm_mappings,
                unsigned int num_elements,
                unsigned int sm_range_min,
                unsigned int sm_range_max,
                unsigned int histo_height,
                unsigned int histo_width,
                unsigned int* global_subhisto, size_t bytes_global_subhisto,
                unsigned int* global_histo, size_t bytes_global_histo,
                unsigned int* global_overflow, size_t bytes_global_overflow,
                int blockx, int blocky,
                int gridx, int gridy) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(1, sm_mappings, 3, global_subhisto, global_histo, global_overflow);

    void* AllocationNode = __visc__createNode(MainAllocation);
    void* MainLeafNode = __visc__createNode2D(MainLeaf, blockx, blocky);

    // Bind Inputs
    __visc__bindIn(AllocationNode, 5, 0, 0);  // Bind block

    __visc__bindIn(MainLeafNode, 0, 0, 0); // Bind sm_mappings
    __visc__bindIn(MainLeafNode, 1, 1, 0); // Bind bytes_sm_mappings
    __visc__bindIn(MainLeafNode, 2, 2, 0); // Bind num_elements
    __visc__bindIn(MainLeafNode, 3, 3, 0); // Bind sm_range_min
    __visc__bindIn(MainLeafNode, 4, 4, 0); // Bind sm_range_max
    __visc__bindIn(MainLeafNode, 5, 5, 0); // Bind histo_height
    __visc__bindIn(MainLeafNode, 6, 6, 0); // Bind histo_width
    __visc__bindIn(MainLeafNode, 7, 7, 0); // Bind global_subhisto
    __visc__bindIn(MainLeafNode, 8, 8, 0); // Bind bytes_global_subhisto
    __visc__bindIn(MainLeafNode, 9, 9, 0); // Bind global_histo
    __visc__bindIn(MainLeafNode, 10, 10, 0); // Bind bytes_global_histo
    __visc__bindIn(MainLeafNode, 11, 11, 0); // Bind overflow
    __visc__bindIn(MainLeafNode, 12, 12, 0); // Bind bytes_overflow

    // Create Edges
    __visc__edge(AllocationNode, MainLeafNode, 1, 0, 13, 0); // Edge sub_histo
    __visc__edge(AllocationNode, MainLeafNode, 1, 1, 14, 0); // Edge bytes_sub_histo

}

void MainRoot (uchar4* sm_mappings, size_t bytes_sm_mappings,
               unsigned int num_elements,
               unsigned int sm_range_min,
               unsigned int sm_range_max,
               unsigned int histo_height,
               unsigned int histo_width,
               unsigned int* global_subhisto, size_t bytes_global_subhisto,
               unsigned int* global_histo, size_t bytes_global_histo,
               unsigned int* global_overflow, size_t bytes_global_overflow,
               int blockx, int blocky,
               int gridx, int gridy) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(1, sm_mappings, 3, global_subhisto, global_histo, global_overflow);

    void* MainBlockNode = __visc__createNode2D(MainBlock, gridx, gridy);

    // Bind Inputs
    __visc__bindIn(MainBlockNode, 0, 0, 0); // Bind sm_mappings
    __visc__bindIn(MainBlockNode, 1, 1, 0); // Bind bytes_sm_mappings
    __visc__bindIn(MainBlockNode, 2, 2, 0); // Bind num_elements
    __visc__bindIn(MainBlockNode, 3, 3, 0); // Bind sm_range_min
    __visc__bindIn(MainBlockNode, 4, 4, 0); // Bind sm_range_max
    __visc__bindIn(MainBlockNode, 5, 5, 0); // Bind histo_height
    __visc__bindIn(MainBlockNode, 6, 6, 0); // Bind histo_width
    __visc__bindIn(MainBlockNode, 7, 7, 0); // Bind global_subhisto
    __visc__bindIn(MainBlockNode, 8, 8, 0); // Bind bytes_global_subhisto
    __visc__bindIn(MainBlockNode, 9, 9, 0); // Bind global_histo
    __visc__bindIn(MainBlockNode, 10, 10, 0); // Bind bytes_global_histo
    __visc__bindIn(MainBlockNode, 11, 11, 0); // Bind overflow
    __visc__bindIn(MainBlockNode, 12, 12, 0); // Bind bytes_overflow
    __visc__bindIn(MainBlockNode, 13, 13, 0); // Bind blockx
    __visc__bindIn(MainBlockNode, 14, 14, 0); // Bind blocky
}

typedef struct __attribute__((__packed__)) {
    unsigned char* sm_mappings;
    size_t bytes_sm_mappings;
    unsigned int num_elements;
    unsigned int sm_range_min;
    unsigned int sm_range_max;
    unsigned int histo_height;
    unsigned int histo_width;
    unsigned int* global_subhisto;
    size_t bytes_global_subhisto;
    unsigned int* global_histo;
    size_t bytes_global_histo;
    unsigned int* global_overflow;
    size_t bytes_global_overflow;
    int blockx;
    int blocky;
    int gridx;
    int gridy;
}
MainRootIn;


void MainPackData(MainRootIn* args,
                  unsigned char* sm_mappings, size_t bytes_sm_mappings,
                  unsigned int num_elements,
                  unsigned int sm_range_min,
                  unsigned int sm_range_max,
                  unsigned int histo_height,
                  unsigned int histo_width,
                  unsigned int* global_subhisto, size_t bytes_global_subhisto,
                  unsigned int* global_histo, size_t bytes_global_histo,
                  unsigned int* global_overflow, size_t bytes_global_overflow,
                  int blockx, int blocky,
                  int gridx, int gridy
                 ) {
    args->sm_mappings = sm_mappings;
    args->bytes_sm_mappings = bytes_sm_mappings;
    args->num_elements = num_elements;
    args->sm_range_min = sm_range_min;
    args->sm_range_max = sm_range_max;
    args->histo_height = histo_height;
    args->histo_width = histo_width;
    args->global_subhisto = global_subhisto;
    args->bytes_global_subhisto = bytes_global_subhisto;
    args->global_histo = global_histo;
    args->bytes_global_histo = bytes_global_histo;
    args->global_overflow = global_overflow;
    args->bytes_global_overflow = bytes_global_overflow;
    args->blockx = blockx;
    args->blocky = blocky;
    args->gridx = gridx;
    args->gridy = gridy;
}

// Final
typedef struct __attribute__((__packed__)) {
    unsigned short x;
    unsigned short y;
    unsigned short z;
    unsigned short w;
}
ushort4;

typedef struct __attribute__((__packed__)) {
    unsigned int x;
    unsigned int y;
    unsigned int z;
    unsigned int w;
}
uint4;
/* Combine all the sub-histogram results into one final histogram */
void FinalLeaf (
    unsigned int sm_range_min, 
    unsigned int sm_range_max,
    unsigned int histo_height, 
    unsigned int histo_width,
    unsigned int *global_subhisto, size_t bytes_global_subhisto,
    unsigned int *global_histo, size_t bytes_global_histo,
    unsigned int *global_overflow, size_t bytes_global_overflow,
    unsigned int *final_histo, size_t bytes_final_histo) //final output
{
    __visc__hint(visc::DEVICE);
    __visc__attributes(3, global_subhisto, global_histo, global_overflow, 1, final_histo);

    void* thisNode = __visc__getNode();
    void* parentNode = __visc__getParentNode(thisNode);
    int lx = __visc__getNodeInstanceID_x(thisNode);
    int gx = __visc__getNodeInstanceID_x(parentNode);
    int dimx = __visc__getNumNodeInstances_x(thisNode);
    int gdimx = __visc__getNumNodeInstances_x(parentNode);


  
    unsigned int blockDimx = dimx;
    unsigned int gridDimx = gdimx;
    unsigned int start_offset = lx + gx * blockDimx;
    const ushort4 zero_short  = {0, 0, 0, 0};
    const uint4 zero_int      = {0, 0, 0, 0};

    unsigned int size_low_histo = sm_range_min * BINS_PER_BLOCK;
    unsigned int size_mid_histo = (sm_range_max - sm_range_min +1) * BINS_PER_BLOCK;

    /* Clear lower region of global histogram */
    for (unsigned int i = start_offset; i < size_low_histo/4; i += gridDimx * blockDimx)
    {
        ushort4 global_histo_data = ((ushort4*)global_histo)[i];
        ((ushort4*)global_histo)[i] = zero_short;
        

        global_histo_data.x = min (global_histo_data.x, (ushort) 255);
        global_histo_data.y = min (global_histo_data.y, (ushort) 255);
        global_histo_data.z = min (global_histo_data.z, (ushort) 255);
        global_histo_data.w = min (global_histo_data.w, (ushort) 255);

        uchar4 final_histo_data = (uchar4) {
            (unsigned char) global_histo_data.x,
            (unsigned char) global_histo_data.y,
            (unsigned char) global_histo_data.z,
            (unsigned char) global_histo_data.w
        };

        ((uchar4*)final_histo)[i] = final_histo_data;
    }

    /* Clear the middle region of the overflow buffer */
    for (unsigned int i = (size_low_histo/4) + start_offset; i < (size_low_histo+size_mid_histo)/4; i += gridDimx * blockDimx)
    {
        uint4 global_histo_data = ((uint4*)global_overflow)[i];
        //((uint4*)global_overflow)[i] = zero_int;
        global_overflow[i*4] = 0;
        global_overflow[i*4+1] = 0;
        global_overflow[i*4+2] = 0;
        global_overflow[i*4+3] = 0;

        uint4 internal_histo_data = (uint4){
            global_histo_data.x,
            global_histo_data.y,
            global_histo_data.z,
            global_histo_data.w
        };

        #pragma unroll
        for (int j = 0; j < BLOCK_X; j++)
        {
            unsigned int bin4in = ((unsigned int*)global_subhisto)[i + j * histo_height * histo_width / 4];
            internal_histo_data.x += (bin4in >>  0) & 0xFF;
            internal_histo_data.y += (bin4in >>  8) & 0xFF;
            internal_histo_data.z += (bin4in >> 16) & 0xFF;
            internal_histo_data.w += (bin4in >> 24) & 0xFF;
        }

        internal_histo_data.x = min (internal_histo_data.x, (uint) 255);
        internal_histo_data.y = min (internal_histo_data.y, (uint) 255);
        internal_histo_data.z = min (internal_histo_data.z, (uint) 255);
        internal_histo_data.w = min (internal_histo_data.w, (uint) 255);

        uchar4 final_histo_data = (uchar4) {
            (unsigned char) internal_histo_data.x,
            (unsigned char) internal_histo_data.y,
            (unsigned char) internal_histo_data.z,
            (unsigned char) internal_histo_data.w
        };

        ((uchar4*)final_histo)[i] = final_histo_data;
    }

    /* Clear the upper region of global histogram */
    for (unsigned int i = ((size_low_histo+size_mid_histo)/4) + start_offset; i < (histo_height*histo_width)/4; i += gridDimx * blockDimx)
    {
        ushort4 global_histo_data = ((ushort4*)global_histo)[i];
        ((ushort4*)global_histo)[i] = zero_short;

        global_histo_data.x = min (global_histo_data.x, (ushort) 255);
        global_histo_data.y = min (global_histo_data.y, (ushort) 255);
        global_histo_data.z = min (global_histo_data.z, (ushort) 255);
        global_histo_data.w = min (global_histo_data.w, (ushort) 255);

        uchar4 final_histo_data = (uchar4) {
            (unsigned char) global_histo_data.x,
            (unsigned char) global_histo_data.y,
            (unsigned char) global_histo_data.z,
            (unsigned char) global_histo_data.w
        };

        ((uchar4*)final_histo)[i] = final_histo_data;
    }
}

void FinalBlock(
    unsigned int sm_range_min, 
    unsigned int sm_range_max,
    unsigned int histo_height, 
    unsigned int histo_width,
    unsigned int *global_subhisto, size_t bytes_global_subhisto,
    unsigned int *global_histo, size_t bytes_global_histo,
    unsigned int *global_overflow, size_t bytes_global_overflow,
    unsigned int *final_histo, size_t bytes_final_histo, //final output
    int block) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(3, global_subhisto, global_histo, global_overflow, 1, final_histo);
    
    void* FinalLeafNode = __visc__createNode1D(FinalLeaf, block);

    // Bind Inputs
    __visc__bindIn(FinalLeafNode, 0, 0, 0); // Bind sm_range_min
    __visc__bindIn(FinalLeafNode, 1, 1, 0); // Bind sm_range_max
    __visc__bindIn(FinalLeafNode, 2, 2, 0); // Bind histo_height
    __visc__bindIn(FinalLeafNode, 3, 3, 0); // Bind histo_width
    __visc__bindIn(FinalLeafNode, 4, 4, 0); // Bind global_subhisto
    __visc__bindIn(FinalLeafNode, 5, 5, 0); // Bind bytes_global_subhisto
    __visc__bindIn(FinalLeafNode, 6, 6, 0); // Bind global_histo
    __visc__bindIn(FinalLeafNode, 7, 7, 0); // Bind bytes_global_histo
    __visc__bindIn(FinalLeafNode, 8, 8, 0); // Bind global_overflow
    __visc__bindIn(FinalLeafNode, 9, 9, 0); // Bind bytes_global_overflow
    __visc__bindIn(FinalLeafNode, 10, 10, 0); // Bind final_histo
    __visc__bindIn(FinalLeafNode, 11, 11, 0); // Bind bytes_final_histo
}

void FinalRoot(
    unsigned int sm_range_min, 
    unsigned int sm_range_max,
    unsigned int histo_height, 
    unsigned int histo_width,
    unsigned int *global_subhisto, size_t bytes_global_subhisto,
    unsigned int *global_histo, size_t bytes_global_histo,
    unsigned int *global_overflow, size_t bytes_global_overflow,
    unsigned int *final_histo, size_t bytes_final_histo, //final output
    int block,
    int grid) {

    __visc__hint(visc::CPU_TARGET);
    __visc__attributes(3, global_subhisto, global_histo, global_overflow, 1, final_histo);
    
    void* FinalBlockNode = __visc__createNode1D(FinalBlock, grid);

    // Bind Inputs
    __visc__bindIn(FinalBlockNode, 0, 0, 0); // Bind sm_range_min
    __visc__bindIn(FinalBlockNode, 1, 1, 0); // Bind sm_range_max
    __visc__bindIn(FinalBlockNode, 2, 2, 0); // Bind histo_height
    __visc__bindIn(FinalBlockNode, 3, 3, 0); // Bind histo_width
    __visc__bindIn(FinalBlockNode, 4, 4, 0); // Bind global_subhisto
    __visc__bindIn(FinalBlockNode, 5, 5, 0); // Bind bytes_global_subhisto
    __visc__bindIn(FinalBlockNode, 6, 6, 0); // Bind global_histo
    __visc__bindIn(FinalBlockNode, 7, 7, 0); // Bind bytes_global_histo
    __visc__bindIn(FinalBlockNode, 8, 8, 0); // Bind global_overflow
    __visc__bindIn(FinalBlockNode, 9, 9, 0); // Bind bytes_global_overflow
    __visc__bindIn(FinalBlockNode, 10, 10, 0); // Bind final_histo
    __visc__bindIn(FinalBlockNode, 11, 11, 0); // Bind bytes_final_histo
    __visc__bindIn(FinalBlockNode, 12, 12, 0); // Bind block
}

typedef struct __attribute__((__packed__)) {
    unsigned int sm_range_min; 
    unsigned int sm_range_max;
    unsigned int histo_height; 
    unsigned int histo_width;
    unsigned int *global_subhisto;
    size_t bytes_global_subhisto;
    unsigned int *global_histo;
    size_t bytes_global_histo;
    unsigned int *global_overflow;
    size_t bytes_global_overflow;
    unsigned int *final_histo;
    size_t bytes_final_histo; //final output
    int block;
    int grid;
}
FinalRootIn;

void FinalPackData( FinalRootIn* args,
    unsigned int sm_range_min, 
    unsigned int sm_range_max,
    unsigned int histo_height, 
    unsigned int histo_width,
    unsigned int *global_subhisto, size_t bytes_global_subhisto,
    unsigned int *global_histo, size_t bytes_global_histo,
    unsigned int *global_overflow, size_t bytes_global_overflow,
    unsigned int *final_histo, size_t bytes_final_histo, //final output
    int block,
    int grid) {
  args->sm_range_min = sm_range_min; 
  args->sm_range_max = sm_range_max;
  args->histo_height = histo_height; 
  args->histo_width = histo_width;
  args->global_subhisto = global_subhisto;
  args->bytes_global_subhisto = bytes_global_subhisto;
  args->global_histo = global_histo;
  args->bytes_global_histo = bytes_global_histo;
  args->global_overflow = global_overflow;
  args->bytes_global_overflow = bytes_global_overflow;
  args->final_histo = final_histo;
  args->bytes_final_histo = bytes_final_histo; //final output
  args->block = block;
  args->grid = grid;
}


int main(int argc, char* argv[]) {
    struct pb_TimerSet timers;
    struct pb_Parameters *parameters;

    parameters = pb_ReadParameters(&argc, argv);
    if (!parameters)
        return -1;

    if(!parameters->inpFiles[0]) {
        fputs("Input file expected\n", stderr);
        return -1;
    }

    char viscOverhead[] = "viscOverhead";
    char prescans[] = "PreScanKernel";
    char postpremems[] = "PostPreMems";
    char memsets[] = "Memsets";
    char intermediates[] = "IntermediatesKernel";
    char mains[] = "MainKernel";
    char finals[] = "FinalKernel";

    int numIterations;
    if (argc >= 2) {
        numIterations = atoi(argv[1]);
    } else {
        fputs("Expected at least one command line argument\n", stderr);
        return -1;
    }

    unsigned int img_width, img_height;
    unsigned int histo_width, histo_height;
    unsigned int nThreads;

    FILE* f = fopen(parameters->inpFiles[0],"rb");
    int result = 0;

    result += fread(&img_width,    sizeof(unsigned int), 1, f);
    result += fread(&img_height,   sizeof(unsigned int), 1, f);
    result += fread(&histo_width,  sizeof(unsigned int), 1, f);
    result += fread(&histo_height, sizeof(unsigned int), 1, f);

    if (result != 4) {
        fputs("Error reading input and output dimensions from file\n", stderr);
        return -1;
    }

    unsigned int* img = (unsigned int*) malloc (img_width*img_height*sizeof(unsigned int));
    unsigned char* histo = (unsigned char*) calloc (histo_width*histo_height, sizeof(unsigned char));

    result = fread(img, sizeof(unsigned int), img_width*img_height, f);

    fclose(f);

    if (result != img_width*img_height) {
        fputs("Error reading input array from file\n", stderr);
        return -1;
    }

    pb_InitializeTimerSet(&timers);
    __visc__init();
    pb_AddSubTimer(&timers, viscOverhead, visc_TimerID_COMPUTATION);
    pb_AddSubTimer(&timers, prescans, visc_TimerID_COMPUTATION);
    pb_AddSubTimer(&timers, postpremems, visc_TimerID_COMPUTATION);
    pb_AddSubTimer(&timers, memsets, visc_TimerID_COMPUTATION);
    pb_AddSubTimer(&timers, intermediates, visc_TimerID_COMPUTATION);
    pb_AddSubTimer(&timers, mains, visc_TimerID_COMPUTATION);
    pb_AddSubTimer(&timers, finals, visc_TimerID_COMPUTATION);


    void* histo_prescan_dfg;
    void* histo_intermediates_dfg;
    void* histo_main_dfg;
    void* histo_final_kernel;

    int even_width = ((img_width+1)/2)*2;

    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    size_t bytes_input = even_width*(((img_height+UNROLL-1)/UNROLL)*UNROLL)*sizeof(unsigned int);
    size_t bytes_ranges = 2*sizeof(unsigned int);
    size_t bytes_sm_mappings = img_width*img_height*4*sizeof(unsigned char);
    size_t bytes_global_subhisto = img_width*histo_height*sizeof(unsigned int);
    size_t bytes_global_histo = img_width*histo_height*sizeof(unsigned short);
    size_t bytes_global_overflow = img_width*histo_height*sizeof(unsigned int);
    size_t bytes_final_histo = img_width*histo_height*sizeof(unsigned char);

    unsigned* input = (unsigned*) malloc(bytes_input);
    unsigned* ranges = (unsigned*) malloc(bytes_ranges);
    unsigned char* sm_mappings = (unsigned char*) malloc(bytes_sm_mappings);
    unsigned int* global_subhisto = (unsigned int*) malloc(bytes_global_subhisto);
    unsigned short* global_histo = (unsigned short*) malloc(bytes_global_histo);
    unsigned int* global_overflow = (unsigned int*) malloc(bytes_global_overflow);
    unsigned char* final_histo = (unsigned char*) malloc(bytes_final_histo);

    pb_SwitchToTimer(&timers, visc_TimerID_MEM_TRACK);
    llvm_visc_track_mem(input, bytes_input);
    llvm_visc_track_mem(ranges, bytes_ranges);
    llvm_visc_track_mem(sm_mappings, bytes_sm_mappings);
    llvm_visc_track_mem(global_histo, bytes_global_histo);
    llvm_visc_track_mem(global_overflow, bytes_global_overflow);
    llvm_visc_track_mem(final_histo, bytes_final_histo);

    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
    switch (KB) {
    case 48:
        nThreads = 1024;
        break;
    case 24:
        nThreads = 768;
        break;
    default:
        nThreads = 512;
        break;
    }


    // FIXME: How to do padding without these many copies
    for (int y=0; y < img_height; y++) {
        memcpy(&input[y*even_width], &img[y*img_width], img_width*sizeof(unsigned int));
    }

    int prescan_localWS = PRESCAN_THREADS;
    int prescan_globalWS = PRESCAN_BLOCKS_X;
    int inter_localWS = (img_width+1)/2;
    int inter_globalWS = ((img_height + UNROLL-1)/UNROLL);
    int main_localWS[2] = {(int)nThreads, 1};
    int main_globalWS[2];
    main_globalWS[0] = BLOCK_X;
    int final_localWS = 512;
    int final_globalWS = BLOCK_X*3;

    pb_SwitchToTimer(&timers, visc_TimerID_ARG_PACK);

    unsigned int img_dim = img_height*img_width;
    PrescanRootIn* argsP = (PrescanRootIn*) malloc(sizeof(PrescanRootIn));
    PrescanPackData(argsP,
                    input, bytes_input,
                    img_dim,
                    ranges, bytes_ranges,
                    prescan_localWS, prescan_globalWS);

    unsigned int half_width = (img_width+1)/2;

    IntermediatesRootIn* argsI = (IntermediatesRootIn*) malloc(sizeof(IntermediatesRootIn));
    IntermediatesPackData(argsI,
                          input, bytes_input,
                          img_height,
                          img_width,
                          half_width,
                          sm_mappings, bytes_sm_mappings,
                          inter_localWS, inter_globalWS);

    MainRootIn* argsM = (MainRootIn*) malloc (sizeof(MainRootIn));
    MainPackData( argsM,
                  sm_mappings, bytes_sm_mappings,
                  img_dim,
                  ranges[0], // This would actually be set inside loop
                  ranges[1], // This would actually be set inside loop
                  histo_height,
                  histo_width,
                  global_subhisto, bytes_global_subhisto,
                  (unsigned int*)global_histo, bytes_global_histo,
                  global_overflow, bytes_global_overflow,
                  main_localWS[0], main_localWS[1],
                  main_globalWS[0], main_globalWS[1]
                  );


    FinalRootIn* argsF = (FinalRootIn*) malloc(sizeof(FinalRootIn));
    FinalPackData(argsF,
                  ranges[0], // This would actually be set inside loop
                  ranges[1], // This would actually be set inside loop
                  histo_height,
                  histo_width,
                  global_subhisto, bytes_global_subhisto,
                  (unsigned int*)global_histo, bytes_global_histo,
                  global_overflow, bytes_global_overflow,
                  (unsigned int*)final_histo, bytes_final_histo,
                  final_localWS, final_globalWS
                  );


    pb_SwitchToTimer(&timers, visc_TimerID_COMPUTATION);

#ifndef debug
    char inputf[] = "input-visc";
    char sm_mappingsf[] = "sm_mappings-visc";
    char global_subhistof[] = "global_subhisto-visc";
    char global_histof[] = "global_histo-visc";
    char global_overflowf[] = "global_overflow-visc";
    numIterations = 1;
#endif

        //memset(global_subhisto, 0, img_width*histo_height*sizeof(unsigned int));
    for (int iter = 0; iter < numIterations; iter++) {
        llvm_visc_track_mem(global_subhisto, bytes_global_subhisto);
        pb_SwitchToSubTimer(&timers, postpremems , visc_TimerID_COMPUTATION);
        //llvm_visc_request_mem(ranges, bytes_ranges);
        //llvm_visc_request_mem(input, bytes_input);
        ranges[0] = UINT32_MAX;
        ranges[1] = 0;

        // how about something like
        // __global__ unsigned int ranges[2];
        // ...kernel
        // __shared__ unsigned int s_ranges[2];
        // if (threadIdx.x == 0) {s_ranges[0] = ranges[0]; s_ranges[1] = ranges[1];}
        // __syncthreads();

        // Although then removing the blocking cudaMemcpy's might cause something about
        // concurrent kernel execution.
        // If kernel launches are synchronous, then how can 2 kernels run concurrently? different host threads?

#ifndef debug        
        dump_histo_img((unsigned char*)input, 1, bytes_input, inputf);
#endif
        pb_SwitchToSubTimer(&timers, prescans , visc_TimerID_COMPUTATION);
        void* PrescanDFG = __visc__launch(0, PrescanRoot, (void*)argsP);
        __visc__wait(PrescanDFG);

        pb_SwitchToSubTimer(&timers, postpremems , visc_TimerID_COMPUTATION);
        
        llvm_visc_request_mem(ranges, bytes_ranges);

        //printf("Range: (%d, %d)\n", ranges[0], ranges[1]);
#ifndef debug
        printf("Range: (%d, %d)\n", ranges[0], ranges[1]);
#endif
        // Requesting only so that we can write it to zero. IT should not be
        // copied from device to host
        llvm_visc_request_mem(global_subhisto, bytes_global_subhisto);
        // Set global_subhisto to zero
        memset(global_subhisto, 0, bytes_global_subhisto);
        //pb_SwitchToSubTimer(&timers, memsets , visc_TimerID_COMPUTATION);

        pb_SwitchToSubTimer(&timers, intermediates, visc_TimerID_COMPUTATION);

        void* IntermediatesDFG = __visc__launch(0, IntermediatesRoot, (void*)argsI);
        __visc__wait(IntermediatesDFG);

        pb_SwitchToSubTimer(&timers, viscOverhead, visc_TimerID_COMPUTATION);
#ifndef debug
        llvm_visc_request_mem(sm_mappings, bytes_sm_mappings);
        dump_histo_img(sm_mappings, 1, bytes_sm_mappings, sm_mappingsf);
#endif
        argsM->gridy = ranges[1]-ranges[0]+1;

        argsM->sm_range_min = ranges[0];
        argsM->sm_range_max = ranges[1];
        argsF->sm_range_min = ranges[0];
        argsF->sm_range_max = ranges[1];
        
        pb_SwitchToSubTimer(&timers, mains, visc_TimerID_COMPUTATION);

        void* MainDFG = __visc__launch(0, MainRoot, (void*) argsM);
        __visc__wait(MainDFG);
#ifndef debug
        llvm_visc_request_mem(global_subhisto, bytes_global_subhisto);
        dump_histo_img((unsigned char*)global_subhisto, 1, bytes_global_subhisto, global_subhistof);

        llvm_visc_request_mem(global_histo, bytes_global_histo);
        dump_histo_img((unsigned char*)global_histo, 1, bytes_global_histo, global_histof);

        llvm_visc_request_mem(global_overflow, bytes_global_overflow);
        dump_histo_img((unsigned char*)global_overflow, 1, bytes_global_overflow, global_overflowf);
#endif
        pb_SwitchToSubTimer(&timers, finals, visc_TimerID_COMPUTATION);

        void* FinalDFG = __visc__launch(0, FinalRoot, (void*) argsF);
        __visc__wait(FinalDFG);

        pb_SwitchToSubTimer(&timers, viscOverhead, visc_TimerID_COMPUTATION);
        llvm_visc_untrack_mem(global_subhisto);
    }

    pb_SwitchToTimer(&timers, pb_TimerID_COPY);
    llvm_visc_request_mem(final_histo, bytes_final_histo);

    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

    llvm_visc_untrack_mem(input);
    llvm_visc_untrack_mem(ranges);
    llvm_visc_untrack_mem(sm_mappings);
    llvm_visc_untrack_mem(global_subhisto);
    llvm_visc_untrack_mem(global_histo);
    llvm_visc_untrack_mem(global_overflow);
    llvm_visc_untrack_mem(final_histo);

    pb_SwitchToTimer(&timers, pb_TimerID_NONE);
    pb_PrintTimerSet(&timers);
    __visc__cleanup();

    if (parameters->outFile) {
        dump_histo_img(final_histo, histo_height, histo_width, parameters->outFile);
    }

    //pb_SwitchToTimer(&timers, visc_TimerID_MEM_FREE);

    free(img);
    free(input);
    free(ranges);
    free(sm_mappings);
    free(global_subhisto);
    free(global_histo);
    free(global_overflow);
    free(final_histo);

    //pb_SwitchToTimer(&timers, pb_TimerID_NONE);

    printf("\n");
    pb_FreeParameters(parameters);

    pb_DestroyTimerSet(&timers);

    return 0;
}
