/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/*
 * Main entry of dense matrix-matrix multiplication kernel
 */

#include "opencv2/opencv.hpp"
#include "opencv2/ocl/ocl.hpp"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <malloc.h>
#include <vector>
#include <iostream>
#include <cassert>
#include <parboil.h>
#include <visc.h>

// I/O routines
extern bool readColMajorMatrixFile(const char *fn, int &nr_row, int &nr_col, std::vector<float>&v);
extern bool writeColMajorMatrixFile(const char *fn, int, int, std::vector<float>&);
extern char* readFile(const char*);

// Definitions of sizes for edge detection kernels

#define MIN_BR 0.0f
#define MAX_BR 1.0f

// Code needs to be changed for this to vary
#define SZB 3

#define REDUCTION_TILE_SZ 1024

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

extern "C" {
//struct __attribute__((__packed__)) OutStruct {
  //int m;
  //int n;
  //size_t bytesB;
  //size_t bytesOut;
//};

struct __attribute__((__packed__)) InStruct {
  float* I ;
  size_t bytesI;
  float* B;
  size_t bytesB;
  float* L;
  size_t bytesL;
  float* S;
  size_t bytesS;
  float* G;
  size_t bytesG;
  float* E;
  size_t bytesE;
  int m;
  int n;
};


void packData(struct InStruct* args, float* I, size_t bytesI,
                                     float* B, size_t bytesB,
                                     float* L, size_t bytesL,
                                     float* S, size_t bytesS,
                                     float* G, size_t bytesG,
                                     float* E, size_t bytesE,
                                     int m, int n) {
  args->I = I;
  args->bytesI = bytesI;
  args->B = B;
  args->bytesB = bytesB;
  args->L = L;
  args->bytesL = bytesL;
  args->S = S;
  args->bytesS = bytesS;
  args->G = G;
  args->bytesG = bytesG;
  args->E = E;
  args->bytesE = bytesE;
  args->m = m;
  args->n = n;
}

/* Compute a non-linear laplacian estimate of input image I of size m x n */
/*
I    : imput image
m, n : dimensions
B    : structural element for dilation - erosion ([0 1 0; 1 1 1; 0 1 0])
L    : output (laplacian of the image)
Need 2D grid, a thread per pixel
*/
void laplacianEstimate(float *I, size_t bytesI,
                            float *B, size_t bytesB,
                            float *L, size_t bytesL,
                            int m, int n) {

  __visc__hint(visc::CPU_TARGET);
  //__visc__hint(visc::GPU_TARGET);
  __visc__attributes(2, I, B, 1, L);
  // 3x3 image area
  float imageArea[SZB][SZB];

  //int gx = get_global_id(0);
  //int gy = get_global_id(1);
  void* thisNode = __visc__getNode();
  int gx = __visc__getNodeInstanceID_x(thisNode);
  int gy = __visc__getNodeInstanceID_y(thisNode);
  //if(gx == 0 && gy == 0)
    //std::cout << "Entered laplacian\n";

  int i, j;

  if ((gx < n) && (gy < m)) {
    // Data copy for dilation filter
    imageArea[1][1] = I[gy * n + gx];

    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MIN_BR;
    } else {
      imageArea[1][0] = I[gy * n + gx - 1];
      imageArea[0][0] = (gy > 0) ? I[(gy - 1) * n + gx - 1] : MIN_BR;
      imageArea[2][0] = (gy < m - 1) ? I[(gy + 1) * n + gx - 1] : MIN_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MIN_BR;
    } else {
      imageArea[1][2] = I[gy * n + gx + 1];
      imageArea[0][2] = (gy > 0) ? I[(gy - 1) * n + gx + 1] : MIN_BR;
      imageArea[2][2] = (gy < m - 1) ? I[(gy + 1) * n + gx + 1] : MIN_BR;
    }

    imageArea[0][1] = (gy > 0) ? I[(gy - 1) * n + gx] : MIN_BR;
    imageArea[2][1] = (gy < m - 1) ? I[(gy + 1) * n + gx] : MIN_BR;

    // Compute pixel of dilated image
    float dilatedPixel = MIN_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        dilatedPixel = MAX(dilatedPixel, imageArea[i][j] * B[i*SZB + j]);

    // Data copy for erotion filter - only change the boundary conditions
    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MAX_BR;
    } else {
      if (gy == 0) imageArea[0][0] = MAX_BR;
      if (gy == m-1) imageArea[2][0] = MAX_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MAX_BR;
    } else {
      if (gy == 0) imageArea[0][2] = MAX_BR;
      if (gy == m-1) imageArea[2][2] = MAX_BR;
    }

    if (gy == 0) imageArea[0][1] = MAX_BR;
    if (gy == m-1) imageArea[2][1] = MAX_BR;

    // Compute pixel of eroded image
    float erodedPixel = MAX_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        erodedPixel = MIN(erodedPixel, imageArea[i][j] * B[i*SZB + j]);

    float laplacian = dilatedPixel + erodedPixel - 2 * imageArea[1][1];
    L[gy*n+gx] = laplacian;
  }
  //OutStruct output = {bytesB, bytesL};
  //if(gx == m-1 && gy == n-1)
    //std::cout << "Exit laplacian\n";
  __visc__return(1, m);
}

void WrapperlaplacianEstimate(float *I, size_t bytesI,
                          float *B, size_t bytesB,
                          float *L, size_t bytesL,
                          int m, int n) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, I, B, 1, L);
  void* LNode = __visc__createNodeND(2, laplacianEstimate, (size_t)m, (size_t)n);
  __visc__bindIn(LNode, 0, 0, 0); // Bind I
  __visc__bindIn(LNode, 1, 1, 0); // Bind bytesI
  __visc__bindIn(LNode, 2, 2, 0); // Bind B
  __visc__bindIn(LNode, 3, 3, 0); // Bind bytesB
  __visc__bindIn(LNode, 4, 4, 0); // Bind L
  __visc__bindIn(LNode, 5, 5, 0); // Bind bytesL
  __visc__bindIn(LNode, 6, 6, 0); // Bind m
  __visc__bindIn(LNode, 7, 7, 0); // Bind n

  __visc__bindOut(LNode, 0, 0, 0); // bind output m

}

/* Compute the zero crossings of input image L of size m x n */
/*
L    : imput image (computed Laplacian)
m, n : dimensions
B    : structural element for dilation - erosion ([0 1 0; 1 1 1; 0 1 0])
S    : output (sign of the image)
Need 2D grid, a thread per pixel
 */
void computeZeroCrossings(float *L, size_t bytesL,
                          float *B, size_t bytesB,
                          float *S, size_t bytesS,
                          int m, int n) {
  __visc__hint(visc::CPU_TARGET);
  //__visc__hint(visc::SPIR_TARGET);
  __visc__attributes(2, L, B, 1, S);

  // 3x3 image area
  float imageArea[SZB][SZB];

  //int gx = get_global_id(0);
  //int gy = get_global_id(1);
  void* thisNode = __visc__getNode();
  int gx = __visc__getNodeInstanceID_x(thisNode);
  int gy = __visc__getNodeInstanceID_y(thisNode);
  int i, j;

  //if(gx == 0 && gy == 0)
    //std::cout << "Entered ZC\n";
  if ((gx < n) && (gy < m)) {
    // Data copy for dilation filter
    imageArea[1][1] = L[gy * n + gx];

    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MIN_BR;
    } else {
      imageArea[1][0] = L[gy * n + gx - 1];
      imageArea[0][0] = (gy > 0) ? L[(gy - 1) * n + gx - 1] : MIN_BR;
      imageArea[2][0] = (gy < m - 1) ? L[(gy + 1) * n + gx - 1] : MIN_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MIN_BR;
    } else {
      imageArea[1][2] = L[gy * n + gx + 1];
      imageArea[0][2] = (gy > 0) ? L[(gy - 1) * n + gx + 1] : MIN_BR;
      imageArea[2][2] = (gy < m - 1) ? L[(gy + 1) * n + gx + 1] : MIN_BR;
    }

    imageArea[0][1] = (gy > 0) ? L[(gy - 1) * n + gx] : MIN_BR;
    imageArea[2][1] = (gy < m - 1) ? L[(gy + 1) * n + gx] : MIN_BR;

    // Compute pixel of dilated image
    float dilatedPixel = MIN_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        dilatedPixel = MAX(dilatedPixel, imageArea[i][j] * B[i*SZB + j]);

    // Data copy for erotion filter - only change the boundary conditions
    if (gx == 0) {
      imageArea[0][0] = imageArea[1][0] = imageArea[2][0] = MAX_BR;
    } else {
      if (gy == 0) imageArea[0][0] = MAX_BR;
      if (gy == m-1) imageArea[2][0] = MAX_BR;
    }

    if (gx == n - 1) {
      imageArea[0][2] = imageArea[1][2] = imageArea[2][2] = MAX_BR;
    } else {
      if (gy == 0) imageArea[0][2] = MAX_BR;
      if (gy == m-1) imageArea[2][2] = MAX_BR;
    }

    if (gy == 0) imageArea[0][1] = MAX_BR;
    if (gy == m-1) imageArea[2][1] = MAX_BR;

    // Compute pixel of eroded image
    float erodedPixel = MAX_BR;
    for (i = 0; i < SZB; i++)
      for (j = 0; j < SZB; j++)
        erodedPixel = MIN(erodedPixel, imageArea[i][j] * B[i*SZB + j]);

    float pixelSign = dilatedPixel - erodedPixel;
    S[gy*n+gx] = pixelSign;
  }
  //OutStruct output = {bytesB, bytesS};
  //if(gx == n-1 && gy == n-1)
    //std::cout << "Exit ZC\n";
  // FIXME: Return bytesG instead of L. RZC nodes expects bytesG
  // This is just temporary
  __visc__return(1, m); 
}

void WrapperComputeZeroCrossings(float *L, size_t bytesL,
                          float *B, size_t bytesB,
                          float *S, size_t bytesS,
                          int m, int n) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, L, B, 1, S);
  void* ZCNode = __visc__createNodeND(2, computeZeroCrossings, (size_t)m, (size_t)n);
  __visc__bindIn(ZCNode, 0, 0, 0); // Bind L
  __visc__bindIn(ZCNode, 1, 1, 0); // Bind bytesL
  __visc__bindIn(ZCNode, 2, 2, 0); // Bind B
  __visc__bindIn(ZCNode, 3, 3, 0); // Bind bytesB
  __visc__bindIn(ZCNode, 4, 4, 0); // Bind S
  __visc__bindIn(ZCNode, 5, 5, 0); // Bind bytesS
  __visc__bindIn(ZCNode, 6, 6, 0); // Bind m
  __visc__bindIn(ZCNode, 7, 7, 0); // Bind n

  __visc__bindOut(ZCNode, 0, 0, 0); // bind output m

}

/*
 * S: sign of image
 * G: gradient of (smoothed) image
 * E: output image : edges
 * theta: threshold value
 * maxG: computed maximum graddient of G
 */
void rejectZeroCrossings(float *S, size_t bytesS,
                         float *G, size_t bytesG,
                         float *E, size_t bytesE,
                         //float theta, float maxG,
                         int m, int n) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, S, G, 1, E);

  void* thisNode = __visc__getNode();
  int gx = __visc__getNodeInstanceID_x(thisNode);
  int gy = __visc__getNodeInstanceID_y(thisNode);
  float theta = 1.0;
  float maxG = 10.0;
  if ((gx < n) && (gy < m)) {
    E[gy*n+gx] = ((S[gy*n+gx] > 0.0) && (G[gy*n+gx] > theta*maxG)) ? 1.0 : 0.0 ;
  }
  
  __visc__return(1, m);
       
}

void WrapperRejectZeroCrossings(float *S, size_t bytesS,
                          float *G, size_t bytesG,
                          float *E, size_t bytesE,
                          //float theta, float maxG,
                          int m, int n) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(2, S, G, 1, E);
  void* RZCNode = __visc__createNodeND(2, rejectZeroCrossings, (size_t)m, (size_t)n);
  __visc__bindIn(RZCNode, 0, 0, 0); // Bind S
  __visc__bindIn(RZCNode, 1, 1, 0); // Bind bytesS
  __visc__bindIn(RZCNode, 2, 2, 0); // Bind G
  __visc__bindIn(RZCNode, 3, 3, 0); // Bind bytesG
  __visc__bindIn(RZCNode, 4, 4, 0); // Bind E
  __visc__bindIn(RZCNode, 5, 5, 0); // Bind bytesE
  __visc__bindIn(RZCNode, 6, 6, 0); // Bind m
  __visc__bindIn(RZCNode, 7, 7, 0); // Bind n

  __visc__bindOut(RZCNode, 0, 0, 0); // bind output bytesG

  //return {0};
}



// Pipelined Root node
void edgeDetection(float *I, size_t bytesI,
                          float *B, size_t bytesB,
                          float *L, size_t bytesL,
                          float *S, size_t bytesS,
                          float *G, size_t bytesG,
                          float *E, size_t bytesE,
                          int m, int n) {
  __visc__attributes(2, I, B, 4, L, S, G, E);
  void* LNode = __visc__createNodeND(0, WrapperlaplacianEstimate);
  void* CZCNode = __visc__createNodeND(0, WrapperComputeZeroCrossings);
  void* RZCNode = __visc__createNodeND(0, WrapperRejectZeroCrossings);

  // Laplacian Inputs
  __visc__bindIn(LNode, 0, 0, 1); // Bind I
  __visc__bindIn(LNode, 1, 1, 1); // Bind bytesI
  __visc__bindIn(LNode, 2, 2, 1); // Bind B
  __visc__bindIn(LNode, 3, 3, 1); // Bind bytesB
  __visc__bindIn(LNode, 4, 4, 1); // Bind L
  __visc__bindIn(LNode, 5, 5, 1); // Bind bytesL
  __visc__bindIn(LNode, 12, 6, 1); // Bind m
  __visc__bindIn(LNode, 13, 7, 1); // Bind n

  // Compute ZC Inputs
  __visc__bindIn(CZCNode, 4, 0, 1); // Bind L
  __visc__bindIn(CZCNode, 5, 1, 1); // Get bytesL
  __visc__bindIn(CZCNode, 2, 2, 1); // Bind B
  __visc__bindIn(CZCNode, 3, 3, 1); // Get bytesB
  __visc__bindIn(CZCNode, 6, 4, 1); // Bind S
  __visc__bindIn(CZCNode, 7, 5, 1); // Bind bytesS
  //__visc__bindIn(CZCNode, 12, 6, 1); // Bind m
  __visc__edge(LNode, CZCNode, 1, 0, 6, 1);
  __visc__bindIn(CZCNode, 13, 7, 1); // Bind n

  // Reject ZC Inputs
  __visc__bindIn(RZCNode, 6, 0, 1); // Bind S
  __visc__bindIn(RZCNode, 7, 1, 1); // Get bytesS
  __visc__bindIn(RZCNode, 8, 2, 1); // Bind G
  __visc__bindIn(RZCNode, 9, 3, 1); // Get bytesG
  __visc__bindIn(RZCNode, 10, 4, 1); // Bind E
  __visc__bindIn(RZCNode, 11, 5, 1); // Bind bytesE
  //__visc__bindIn(RZCNode, 12, 6, 1); // Bind m
  __visc__edge(CZCNode, RZCNode, 1, 0, 6, 1);
  __visc__bindIn(RZCNode, 13, 7, 1); // Bind n

  __visc__bindOut(RZCNode, 0, 0, 1); // dummy bind output to get pipeline functionality
}

}
#define NUM_RUNS 5
#define NUM_FRAMES 2000
using namespace cv;

int main (int argc, char *argv[]) {

    struct pb_Parameters *params;
    struct pb_TimerSet timers;

    size_t I_sz, L_sz, S_sz, G_sz, E_sz;
    int matIrow, matIcol;
    std::vector<float> matI;
    std::string window_name = "Edge Map";

    std::cout << "Using OpenCV" << CV_VERSION << "\n";

    /* Read command line. Expect 3 inputs: A, B and B^T
       in column-major layout*/
    params = pb_ReadParameters(&argc, argv);
    if ((params->inpFiles[0] == NULL)
            || (params->inpFiles[1] != NULL))
    {
        fprintf(stderr, "Expecting input image filename\n");
        exit(-1);
    }

    //Read in data 
    //std::cout << "Reading video file: " << params->inpFiles[0] << "\n";
    //VideoCapture cap(params->inpFiles[0]);
    //if(!cap.isOpened()) {
      //std::cout << "Could not open video file" << "\n";
      //return -1;
    //}

    //double count =  cap.get(CV_CAP_PROP_FRAME_COUNT);
    //namedWindow(window_name, CV_WINDOW_AUTOSIZE);

    //Mat src;
    //while(1) {
      //bool success = cap.read(src);
      //if(!success) {
        //std::cout << "Cannot read frame" << "\n";
        //return -1;
      //}
      //imshow(window_name, src);
      //if(waitKey(0) == 27) break;
    //}
    
    Mat src = imread(params->inpFiles[0]);
    if( !src.data )
    {
        return -1;
    }
    /// Create a window
    namedWindow( window_name, CV_WINDOW_AUTOSIZE );
    imshow(window_name, src);
    waitKey(0);
    std::cout << "Image dimension = " << src.size() << "\n";
    /// Convert the image to grayscale
    cvtColor( src, src, CV_BGR2GRAY );

    src.convertTo(src, CV_32F, 1.0/255.0);
    if(src.isContinuous()) {
      matI.assign((float*)src.datastart, (float*)src.dataend);
      std::cout << "Assigning image to float vector\n";
      std::cout << "Vector size = " << matI.size() << "\n";
      matIrow = src.size[0];
      matIcol = src.size[1];
    }
    else {
      std::cout << "Expecting contiguous storage of image in memory!\n";
      exit(-1);
    }

    // Read data from matrix file
    //readColMajorMatrixFile(params->inpFiles[0],
                           //matIrow, matIcol, matI);
    pb_InitializeTimerSet(&timers);
    __visc__init();

    pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
    // copy A to device memory
    I_sz = matIrow*matIcol*sizeof(float);
    L_sz = I_sz;

    // allocate space for C
    S_sz = I_sz;
    G_sz = I_sz;
    E_sz = I_sz;

    // OpenCL memory allocation
    std::vector<float> matL(matIrow*matIcol);
    std::vector<float> matS(matIrow*matIcol);
    std::vector<float> matG(matIrow*matIcol);
    std::vector<float> matE(matIrow*matIcol);

    float B[] = {0, 1, 0, 1, 1, 1, 0, 1, 0};
    size_t bytesB = 9*sizeof(float);

    size_t BlockElements = 16;
    size_t BlockSize = BlockElements * sizeof(float) ;

    // Copy A and B^T into device memory
    pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );

    // TODO: Should be removed
    for(size_t i=0; i<matL.size(); i++) {
        matL[i] = 0.0f;
        matS[i] = 0.0f;
        matG[i] = 0.0f;
        matE[i] = 0.0f;
    }

    pb_SwitchToTimer( &timers, visc_TimerID_COMPUTATION );
    struct InStruct* args = (struct InStruct*)malloc (sizeof(InStruct));
    
    packData(args, &matI.front(), I_sz,
                   B, bytesB,
                   &matL.front(), L_sz,
                   &matS.front(), S_sz,
                   &matG.front(), G_sz,
                   &matE.front(), E_sz,
                   matIrow, matIcol);

    assert(I_sz % BlockSize == 0);
    for(unsigned j=0; j<NUM_RUNS; j++) {
      std::cout << "Run: " << j << "\n";
      void* DFG = __visc__launch(1, edgeDetection, (void*)args);

      llvm_visc_track_mem(&matI[0], I_sz);
      llvm_visc_track_mem(&matL[0], L_sz);
      llvm_visc_track_mem(&matS[0], S_sz);
      llvm_visc_track_mem(&matG[0], G_sz);
      llvm_visc_track_mem(&matE[0], E_sz);
      llvm_visc_track_mem(B, bytesB);
      //packData(args, &matA[0], BlockSize, &matB[i], BlockSize, &matC[i], BlockSize, BlockElements);
      __visc__push(DFG, args);
      __visc__push(DFG, args);
      for(unsigned i=0; i<NUM_FRAMES-2; i++) {
        __visc__push(DFG, args);
        __visc__pop(DFG);
      }
      __visc__pop(DFG);
      __visc__pop(DFG);
      llvm_visc_request_mem(&matS[0], S_sz);
      llvm_visc_untrack_mem(&matI[0]);
      llvm_visc_untrack_mem(&matL[0]);
      llvm_visc_untrack_mem(&matS[0]);
      llvm_visc_untrack_mem(&matG[0]);
      llvm_visc_untrack_mem(&matE[0]);
      llvm_visc_untrack_mem(B);

    //for(unsigned i=0 ; i < I_sz/sizeof(float); i = i + 16) {
      //llvm_visc_track_mem(&matA[i], BlockSize);
      //llvm_visc_track_mem(&matB[i], BlockSize);
      //llvm_visc_track_mem(&matC[i], BlockSize);
      //packData(args, &matA[i], BlockSize, &matB[i], BlockSize, &matC[i], BlockSize, BlockElements);
      //__visc__push(DFG, args);
      //void* output = __visc__pop(DFG);
      //llvm_visc_request_mem(&matC[i], BlockSize);
      //llvm_visc_untrack_mem(&matA[i]);
      //llvm_visc_untrack_mem(&matB[i]);
      //llvm_visc_untrack_mem(&matC[i]);
    //}

      __visc__wait(DFG);
    }
    pb_SwitchToTimer( &timers, pb_TimerID_COPY );
    //llvm_visc_request_mem(&matC.front(), C_sz);

    pb_SwitchToTimer( &timers, visc_TimerID_MEM_UNTRACK );
    //llvm_visc_untrack_mem(&matA.front());
    //llvm_visc_untrack_mem(&matB.front());
    //llvm_visc_untrack_mem(&matC.front());
    pb_SwitchToTimer(&timers, pb_TimerID_NONE);


    pb_PrintTimerSet(&timers);
    __visc__cleanup();

    if (params->outFile) {

        /* Write C to file */
        //pb_SwitchToTimer(&timers, pb_TimerID_IO);
        writeColMajorMatrixFile(params->outFile,
                                matIrow, matIcol, matS);
    }
    Mat output(src.size[0], src.size[1], CV_32F);
    memcpy(output.data, matS.data(), matS.size()*sizeof(float));
    imshow(window_name, output);
    waitKey(0);

    double GPUtime = pb_GetElapsedTime(&(timers.timers[pb_TimerID_KERNEL]));
    std::cout<< "GFLOPs = " << 2.* matIrow * matIcol * matIcol/GPUtime/1e9 << std::endl;
    pb_FreeParameters(params);

    return 0;
}
