
/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <parboil.h>
#include <visc.h>
#include "file.h"
#include "common.h"

static int read_data(float *A0, int nx,int ny,int nz,FILE *fp)
{
    int s=0;
    int i,j,k;
    for(i=0; i<nz; i++)
    {
        for(j=0; j<ny; j++)
        {
            for(k=0; k<nx; k++)
            {
                fread(A0+s,sizeof(float),1,fp);
                s++;
            }
        }
    }
    return 0;
}

void naive_kernel(float c0,float c1, float* A0, float *Anext,int nx,int ny,int nz)
{
    __visc__hint(visc::DEVICE);
    __visc__attributes(2, A0, Anext, 1, Anext);
    int i = get_global_id(0)+1;
    int j = get_global_id(1)+1;
    int k = get_global_id(2)+1;

    if(i<nx-1)
    {
        Anext[Index3D (nx, ny, i, j, k)] = c1 *
                                           ( A0[Index3D (nx, ny, i, j, k + 1)] +
                                             A0[Index3D (nx, ny, i, j, k - 1)] +
                                             A0[Index3D (nx, ny, i, j + 1, k)] +
                                             A0[Index3D (nx, ny, i, j - 1, k)] +
                                             A0[Index3D (nx, ny, i + 1, j, k)] +
                                             A0[Index3D (nx, ny, i - 1, j, k)] )
                                           - A0[Index3D (nx, ny, i, j, k)] * c0;
    }
}

int main(int argc, char** argv) {
    struct pb_TimerSet timers;
    struct pb_Parameters *parameters;

    printf("OpenCL accelerated 7 points stencil codes****\n");
    printf("Author: Li-Wen Chang <lchang20@illinois.edu>\n");
    parameters = pb_ReadParameters(&argc, argv);

    /*pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);*/

    //declaration
    int nx,ny,nz;
    size_t size;
    int iteration;
    float c0=1.0/6.0;
    float c1=1.0/6.0/6.0;

    if (argc<5)
    {
        printf("Usage: probe nx ny nz t\n"
               "nx: the grid size x\n"
               "ny: the grid size y\n"
               "nz: the grid size z\n"
               "t: the iteration time\n");
        return -1;
    }

    nx = atoi(argv[1]);
    if (nx<1)
        return -1;
    ny = atoi(argv[2]);
    if (ny<1)
        return -1;
    nz = atoi(argv[3]);
    if (nz<1)
        return -1;
    iteration = atoi(argv[4]);
    if(iteration<1)
        return -1;

    //host data
    float *h_A0;
    float *h_Anext;

    //load data from files

    size=nx*ny*nz;

    h_A0=(float*)malloc(sizeof(float)*size);
    h_Anext=(float*)malloc(sizeof(float)*size);


    /*pb_SwitchToTimer(&timers, pb_TimerID_IO);*/
    FILE *fp = fopen(parameters->inpFiles[0], "rb");
    read_data(h_A0, nx,ny,nz,fp);
    fclose(fp);

    pb_InitializeTimerSet(&timers);
    __visc__init();

    pb_SwitchToTimer(&timers, visc_TimerID_MEM_TRACK);
    llvm_visc_track_mem(h_A0, sizeof(float)*size);
    llvm_visc_track_mem(h_Anext, sizeof(float)*size);

    pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

    memcpy (h_Anext,h_A0,sizeof(float)*size);


    //only use 1D thread block
    int tx =256;
    long block[3] = {tx,1,1};
    long grid[3] = {(nx-2+tx-1)/tx*tx,ny-2,nz-2};
    //size_t grid[3] = {nx-2,ny-2,nz-2};
    size_t offset[3] = {1,1,1};

    printf("grid(%d, %d, %d), block(%d, %d, %d)\n", grid[0], grid[1], grid[2], block[0], block[1], block[2]);
    //main execution

    int t;
    size_t bytes = size*sizeof(float);
    printf("A[126,1,1] = %f\n", h_A0[Index3D(nx, ny, 126,1,1)]);
    printf("A[125,1,1] = %f\n", h_A0[Index3D(nx, ny, 125,1,1)]);
    for(t=0; t<iteration; t++)
    {
        pb_SwitchToTimer(&timers, pb_TimerID_NONE);
        void* stencilDFG = __visc__node(naive_kernel, 2, 3, block[0], block[1], block[2], grid[0]/block[0], grid[1]/block[1], grid[2]/block[2], 9, (float)c0, (float)c1, h_A0, bytes, h_Anext, bytes, nx, ny, nz, 0);
        __visc__wait(stencilDFG);
        //printf("iteration %d\n",t);
        pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
        float* h_temp = h_A0;
        h_A0 = h_Anext;
        h_Anext = h_temp;

    }


    float* h_temp = h_A0;
    h_A0 = h_Anext;
    h_Anext = h_temp;
    pb_SwitchToTimer(&timers, pb_TimerID_COPY);
    llvm_visc_request_mem(h_Anext, bytes);
    printf("A[126,1,1] = %f\n", h_Anext[Index3D(nx, ny, 126,1,1)]);
    printf("A[125,1,1] = %f\n", h_Anext[Index3D(nx, ny, 125,1,1)]);

    pb_SwitchToTimer(&timers, visc_TimerID_MEM_UNTRACK);

    llvm_visc_untrack_mem(h_A0);
    llvm_visc_untrack_mem(h_Anext);

    pb_SwitchToTimer(&timers, pb_TimerID_NONE);
    pb_PrintTimerSet(&timers);

    __visc__cleanup();

    if (parameters->outFile) {
        /*pb_SwitchToTimer(&timers, pb_TimerID_IO);*/
        outputData(parameters->outFile,h_Anext,nx,ny,nz);

    }
    /*pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);*/
    free(h_A0);
    free(h_Anext);
    pb_FreeParameters(parameters);

    return 0;
}
